import numpy as np
import pyvista as pv
from lxml import etree
import re
import uuid
import time
import threading

from geviewer import geometry


class Parser:
    """Base class for all parsers.
    """

    def __init__(self, filename):
        """Initializes the parser.

        :param filename: The name of the file to parse.
        :type filename: str
        """
        self.filename = filename


    def initialize_template(self, name):
        """Initializes a template for a component.

        :param name: The name of the component.
        :type name: str
        :return: The initialized template.
        :rtype: dict
        """
        return  {'name': name, 'id': str(uuid.uuid4())[-12:], 'shape': '', 'points': [], 'mesh_points': [],\
                 'mesh_inds': [], 'colors': [], 'visible': True, 'scalars': [], 'is_dot': False, \
                 'is_event': False, 'mesh': None, 'has_actor': False, 'children': []}
    
    
    def combine_mesh_arrays(self, points, cells, colors):
        """Combines multiple mesh arrays into a single mesh.

        This function takes lists of points, indices of faces or line segments
        (called cells), and colors, and combines them into a single set of points,
        cells, and colors, adjusting indices appropriately.

        :param points: A list of arrays containing point coordinates.
        :type points: list of numpy.ndarray
        :param cells: A list of lists containing cell indices.
        :type cells: list of list
        :param colors: A list of arrays containing color data.
        :type colors: list of numpy.ndarray
        :return: The combined points, cells, and colors.
        :rtype: tuple
        """

        offsets = np.cumsum([0] + [len(p) for p in points[:-1]]).astype(int)
        points = np.concatenate(points)
        
        for i, cell in enumerate(cells):
            j = 0
            while j < len(cell):
                k = int(cell[j])
                start_idx = j + 1
                end_idx = j + k + 1
                
                cell[start_idx:end_idx] = (np.array(cell[start_idx:end_idx]) + int(offsets[i])).tolist()
                j += k + 1
        
        cells = np.concatenate(cells).astype(int)
        colors = np.concatenate(colors)

        return points, cells, colors


class VRMLParser(Parser):
    """Parser for VRML files.
    """

    def parse_file(self, progress_obj=None):
        """Parses the VRML file and creates the meshes.

        :param progress_obj: The progress object to use.
        :type progress_obj: geviewer.gui.GeProgressBar, optional
        """
        update = 'Reading VRML file...\n'
        if progress_obj:
            if progress_obj.sync_status(update=update): return
        else:
            print(update)
        data = self.read_file(self.filename)
        viewpoint_block, polyline_blocks, marker_blocks, solid_blocks = self.extract_blocks(data, progress_obj=progress_obj)
        self.viewpoint_block = viewpoint_block
        now = time.time()
        polyline_mesh, marker_mesh, solid_mesh = self.create_meshes(polyline_blocks, marker_blocks, solid_blocks, progress_obj=progress_obj)
        component_name = self.filename.split('/')[-1].split('.')[0]
        component = self.initialize_template(component_name)
        names = ['Trajectories', 'Step Markers', 'Geometry']
        for i,mesh in enumerate([polyline_mesh, marker_mesh, solid_mesh]):
            comp = self.initialize_template(names[i])
            comp['mesh'] = mesh
            component['children'].append(comp)

        self.components = component


    def read_file(self, filename):
        """Reads the content of a file.

        :param filename: The path to the file to read.
        :type filename: str
        :return: A single string containing the content of the file.
        :rtype: str
        """
        data = []
        with open(filename, 'r') as f:
            for line in f:
                # don't read comments
                if not line.strip().startswith('#'):
                    data.append(line)
        data = ''.join(data)
        return data
    

    def create_meshes(self, polyline_blocks, marker_blocks, solid_blocks, progress_obj=None):
        """Creates and returns meshes for polylines, markers, and solids.

        This function processes blocks of data for polylines, markers, and solids,
        building corresponding meshes for each.

        :param polyline_blocks: List of blocks containing polyline data.
        :type polyline_blocks: list
        :param marker_blocks: List of blocks containing marker data.
        :type marker_blocks: list
        :param solid_blocks: List of blocks containing solid data.
        :type solid_blocks: list
        :param progress_obj: The progress object to use.
        :type progress_obj: geviewer.gui.GeProgressBar, optional
        :return: The created meshes.
        :rtype: tuple
        """
        update = 'Building meshes...\n'
        if progress_obj:
            if progress_obj.sync_status(update=update): return
        else:
            print(update)

        total = len(polyline_blocks) + len(marker_blocks) + len(solid_blocks)
        if progress_obj:
            progress_obj.reset_progress()
            progress_obj.set_maximum_value(total)
        polyline_mesh = self.build_mesh(polyline_blocks, 'polyline', progress_obj)
        marker_mesh = self.build_markers(marker_blocks, progress_obj)
        solid_mesh = self.build_mesh(solid_blocks, 'solid', progress_obj)

        if progress_obj:
            progress_obj.signal_finished()

        return polyline_mesh, marker_mesh, solid_mesh


    def build_mesh(self, blocks, which, progress_obj=None):
        """Builds a mesh for the given blocks.

        This function processes blocks of data for polylines, markers, and solids,
        building corresponding meshes for each.

        :param blocks: List of blocks containing data.
        :type blocks: list
        :param which: The type of mesh to build.
        :type which: str
        :param progress_obj: The progress object to use.
        :type progress_obj: geviewer.gui.GeProgressBar, optional
        :return: The created mesh.
        :rtype: pyvista.PolyData
        """
        points = [None for i in range(len(blocks))]
        cells = [None for i in range(len(blocks))]
        colors = [None for i in range(len(blocks))]

        if which == 'polyline':
            func = self.process_polyline_block
        elif which == 'solid':
            func = self.process_solid_block

        for i, block in enumerate(blocks):
            points[i], cells[i], color = func(block)
            colors[i] = [color]*len(points[i])
            if progress_obj:
                if progress_obj.sync_status(increment=True): return

        if len(points) == 0:
            return None
        
        points, cells, colors = self.combine_mesh_arrays(points, cells, colors)
        if func==self.process_polyline_block:
            mesh = pv.PolyData(points, lines=cells)
        elif func==self.process_solid_block:
            mesh = pv.PolyData(points, faces=cells)
        mesh.point_data.set_scalars(colors, name='color')

        return mesh


    def build_markers(self, blocks, progress_obj=None):
        """Builds a mesh for the given blocks.

        This function processes blocks of data for markers, creating a mesh for each.

        :param blocks: List of blocks containing data.
        :type blocks: list
        :param progress_obj: The progress object to use.
        :type progress_obj: geviewer.gui.GeProgressBar, optional
        :return: The created mesh.
        :rtype: pyvista.PolyData
        """
        centers = [None for i in range(len(blocks))]
        radii = [None for i in range(len(blocks))]
        colors = [None for i in range(len(blocks))]
        
        for i, block in enumerate(blocks):
            centers[i], radii[i], colors[i] = self.process_marker_block(block)
        if len(centers) == 0:
            return None
        
        mesh = pv.MultiBlock()
        for i in range(len(centers)):
            mesh.append(pv.Sphere(radius=radii[i], center=centers[i]))
            colors[i] = [colors[i]]*mesh[-1].n_points
            if progress_obj:
                if progress_obj.sync_status(increment=True): return

        colors = np.concatenate(colors)
        mesh = mesh.combine()
        mesh.point_data.set_scalars(colors, name='color')

        return mesh


    def extract_blocks(self, file_content, progress_obj=None):
        """Extracts polyline, marker, and solid blocks from the given file content.

        This function processes the provided file content, which is expected to
        be in a text format, and extracts blocks of different types based on
        specific keywords. It separates the blocks into categories: polyline,
        marker, and solid blocks, and also identifies the viewpoint block.

        :param file_content: The content of the file as a single string.
        :type file_content: str
        :param progress_obj: The progress object to use.
        :type progress_obj: geviewer.gui.GeProgressBar, optional
        :return: A tuple containing four elements:
            - The viewpoint block (if found) as a string or `None` if not found.
            - A list of polyline blocks as strings.
            - A list of marker blocks as strings.
            - A list of solid blocks as strings.
        :rtype: tuple
        """
        update = 'Parsing VRML file...\n'
        if progress_obj:
            if progress_obj.sync_status(update=update): return
        else:
            print(update)

        polyline_blocks = []
        marker_blocks = []
        solid_blocks = []
        viewpoint_block = None

        lines = file_content.split('\n')
        block = []
        inside_block = False
        brace_count = 0

        if progress_obj:
            progress_obj.reset_progress()
            progress_obj.set_maximum_value(len(lines))

        for i, line in enumerate(lines):
            stripped_line = line.strip()

            if stripped_line.startswith('Shape') or stripped_line.startswith('Anchor')\
                or stripped_line.startswith('Viewpoint'):
                inside_block = True
                brace_count = 0
            
            if inside_block:
                block.append(line)
                brace_count += line.count('{') - line.count('}')
                
                if brace_count == 0:
                    block_content = '\n'.join(block)
                    
                    if 'IndexedLineSet' in block_content:
                        polyline_blocks.append(block_content)
                    elif 'Sphere' in block_content:
                        marker_blocks.append(block_content)
                    elif 'IndexedFaceSet' in block_content:
                        solid_blocks.append(block_content)
                    elif 'Viewpoint' in block_content:
                        viewpoint_block = block_content

                    block = []
                    inside_block = False

            if progress_obj:
                if progress_obj.sync_status(increment=True): return

        if progress_obj:
            progress_obj.signal_finished()

        return viewpoint_block, polyline_blocks, marker_blocks, solid_blocks


    def process_polyline_block(self, block):
        """Processes a polyline block to create a polyline mesh.

        This function takes a block of polyline data and converts it into a
        PyVista`PolyData` object representing the polyline mesh. It also
        extracts the color information associated with the mesh.

        :param block: The polyline block content as a string.
        :type block: str
        :return: A tuple containing:
            - A `pv.PolyData` object representing the polyline mesh.
            - The color associated with the polyline mesh as a list or array.
        :rtype: tuple
        """
        points, indices, color = self.parse_polyline_block(block)
        lines = []
        for i in range(len(indices) - 1):
            if indices[i] != -1 and indices[i + 1] != -1:
                lines.extend([2, indices[i], indices[i + 1]])
        
        return points, lines, color


    def process_marker_block(self, block):
        """Processes a marker block to create a marker mesh.

        This function takes a block of marker data and creates a spherical
        marker mesh using PyVista. It also extracts the color information
        associated with the marker.

        :param block: The marker block content as a string.
        :type block: str
        :return: A tuple containing:
            - A `pv.Sphere` object representing the marker mesh.
            - The color associated with the marker mesh as a list or array.
        :rtype: tuple
        """
        center, radius, color = self.parse_marker_block(block)

        return center, radius, color


    def process_solid_block(self, block):
        """Processes a solid block to create a solid mesh.

        This function takes a block of solid data and creates a mesh for a
        solid object using PyVista. It also extracts the color information
        associated with the solid.

        :param block: The solid block content as a string.
        :type block: str
        :return: A tuple containing:
            - A `pv.PolyData` object representing the solid mesh.
            - The color associated with the solid mesh as a list or array.
        :rtype: tuple
        """
        points, indices, color = self.parse_solid_block(block)
        faces = []
        current_face = []
        for index in indices:
            if index == -1:
                if len(current_face) == 3:
                    faces.extend([3] + current_face)
                elif len(current_face) == 4:
                    faces.extend([4] + current_face)
                current_face = []
            else:
                current_face.append(index)
        faces = np.array(faces)

        return points, faces, color


    def parse_viewpoint_block(self, block):
        """Parses the viewpoint block to extract the field of view, position,
        and orientation.

        This function extracts the field of view (FOV), position, and orientation
        from a given viewpoint block in a 3D scene description. The FOV is converted
        from radians to degrees.

        :param block: The viewpoint block content as a string.
        :type block: str
        :return: A tuple containing:
            - The field of view in degrees as a float (or None if not found).
            - The position as a list of three floats [x, y, z] (or None if not found).
            - The orientation as a list of four floats [x, y, z, angle] in radians
            (or None if not found).
        :rtype: tuple
        """
        fov = None
        position = None
        orientation = None

        if block is not None:
            fov_match = re.search(r'fieldOfView\s+([\d.]+)', block)
            if fov_match:
                fov = float(fov_match.group(1))*180/np.pi
            
            position_match = re.search(r'position\s+([\d.-]+)\s+([\d.-]+)\s+([\d.-]+)', block)
            if position_match:
                position = [float(position_match.group(1)), float(position_match.group(2)), \
                            float(position_match.group(3))]

            orientation_match = re.search(r'orientation\s+([\d.-]+)\s+([\d.-]+)\s+([\d.-]+)\s+([\d.-]+)', block)
            if orientation_match:
                orientation = [float(orientation_match.group(1)), float(orientation_match.group(2)), \
                            float(orientation_match.group(3)), float(orientation_match.group(4))]
        
        return fov, position, orientation


    def parse_polyline_block(self, block):
        """Parses a polyline block to extract particle track information, including
        coordinates, indices, and color.

        This function processes a block of text representing a polyline in a 3D
        scene description. It extracts the coordinates of the points that define
        the polyline, the indices that describe the lines between these points, 
        and the color associated with the polyline.

        :param block: The polyline block content as a string.
        :type block: str
        :return: A tuple containing:
            - `coords`: An array of shape (N, 3) representing the coordinates of the
            polyline points.
            - `indices`: An array of integers representing the indices that define
            the polyline segments.
            - `color`: An array of four floats representing the RGBA color of the
            polyline, where the alpha is set to 1.
        :rtype: tuple
        """
        coords = []
        coord_inds = []
        color = [1, 1, 1]

        lines = block.split('\n')
        reading_points = False
        reading_indices = False

        for line in lines:
            line = line.strip()
            if line.startswith('point ['):
                reading_points = True
                continue
            elif line.startswith(']'):
                reading_points = False
                reading_indices = False
                continue
            elif line.startswith('coordIndex ['):
                reading_indices = True
                continue
            elif 'diffuseColor' in line:
                color = list(map(float, re.findall(r'[-+]?\d*\.?\d+', line)))
            if reading_points:
                point = line.replace(',', '').split()
                if len(point) == 3:
                    coords.append(list(map(float, point)))
            elif reading_indices:
                indices = line.replace(',', '').split()
                coord_inds.extend(list(map(int, indices)))

        color.append(1.)

        return np.array(coords), np.array(coord_inds), np.array(color)


    def parse_marker_block(self, block):
        """Parses a marker block to extract the position, radius, and color of a marker.

        This function processes a block of text representing a marker in a 3D scene
        description. It extracts the position of the marker, the radius of the marker
        (typically a sphere), and the color of the marker. It also accounts for
        transparency and adjusts the alpha value of the color accordingly.

        :param block: The marker block content as a string.
        :type block: str
        :return: A tuple containing:
            - `coords`: An array of shape (3,) representing the position of the marker
            in 3D space.
            - `radius`: A float representing the radius of the marker.
            - `color`: An array of four floats representing the RGBA color of the marker,
            where alpha is adjusted for transparency.
        :rtype: tuple
        """
        coords = []
        color = [1, 1, 1]
        transparency = 0
        radius = 1

        lines = block.split('\n')

        for line in lines:
            line = line.strip()
            if line.startswith('translation'):
                point = line.split()[1:]
                if len(point) == 3:
                    coords = list(map(float, point))
            elif 'diffuseColor' in line:
                color = list(map(float, re.findall(r'[-+]?\d*\.?\d+', line)))
            elif 'transparency' in line:
                transparency = float(re.findall(r'[-+]?\d*\.?\d+', line)[0])
            elif 'radius' in line:
                radius = float(re.findall(r'[-+]?\d*\.?\d+', line)[0])

        color.append(1. - transparency)

        return np.array(coords), radius, np.array(color)


    def parse_solid_block(self, block):
        """Parses a solid block to extract geometry information for a 3D
        solid object.

        This function processes a block of text representing a solid object
        in a 3D scene description. It extracts the vertex coordinates, the face
        indices that define the geometry of the solid, and the color of the solid.
        The function also handles transparency by adjusting the alpha value in the
        color array.

        :param block: The solid block content as a string.
        :type block: str
        :return: A tuple containing:
            - `coords`: An array of shape (N, 3) where N is the number of vertices,
            representing the vertex coordinates.
            - `coord_inds`: An array of shape (M,) where M is the number of indices,
            representing the indices defining the faces of the solid.
            - `color`: An array of four floats representing the RGBA color of the solid,
            where the alpha value is adjusted for transparency.
        :rtype: tuple
        """
        coords = []
        coord_inds = []
        color = [1, 1, 1]
        transparency = 0

        lines = block.split('\n')
        reading_points = False
        reading_indices = False

        for line in lines:
            line = line.strip()
            if line.startswith('point ['):
                reading_points = True
                continue
            elif line.startswith(']'):
                reading_points = False
                reading_indices = False
                continue
            elif line.startswith('coordIndex ['):
                reading_indices = True
                continue
            elif 'diffuseColor' in line:
                color = list(map(float, re.findall(r'[-+]?\d*\.?\d+', line)))
            elif 'transparency' in line:
                transparency = float(re.findall(r'[-+]?\d*\.?\d+', line)[0])
            if reading_points:
                point = line.replace(',', '').split()
                if len(point) == 3:
                    coords.append(list(map(float, point)))
            elif reading_indices:
                indices = line.replace(',', '').split()
                coord_inds.extend(list(map(int, indices)))

        color.append(1. - transparency)

        return np.array(coords), np.array(coord_inds), np.array(color)


class HepRepParser(Parser):
    """Parser for HepRep files.
    """

    def parse_file(self, progress_obj=None):
        """Parses the HepRep file and creates the meshes.

        :param progress_obj: The progress object to use for the progress bar.
        :type progress_obj: ProgressBar, optional
        """
        if progress_obj:
            print_func = progress_obj.print_update
        else:
            print_func = print

        print_func('Parsing HepRep file...\n')
        if progress_obj:
            progress_obj.reset_progress()
            progress_obj.set_maximum_value(0)

        # approximate total number of elements from number of lines in file
        with open(self.filename, 'r') as f:
            total_elements = sum(1 for _ in f) // 2

        if total_elements > 1e6:
            print_func('This step should take less than {:.0f} seconds.\n'.format(total_elements/1e6))

        self.root = self.parse_xml(self.filename)
        if progress_obj:
            progress_obj.signal_finished()

        component_name = self.filename.split('/')[-1].split('.')[0]
        seed_component = self.initialize_template(component_name)
        self.event_number = 0
        self.num_components = 0
        print_func('Extracting components...\n')

        if progress_obj:
            progress_obj.reset_progress()
            progress_obj.set_maximum_value(total_elements)

        self.populate_meshes(self.root, seed_component, progress_obj=progress_obj)
        self.components = [seed_component]

        if progress_obj:
            progress_obj.signal_finished()
        
        print_func('Building meshes...\n')

        if progress_obj:
            progress_obj.reset_progress()
            progress_obj.set_maximum_value(self.num_components)
        
        self.create_meshes(self.components, progress_obj=progress_obj)
        self.reduce_components(self.components)
        self.build_mesh_objects(self.components)

        if progress_obj:
            progress_obj.signal_finished()


    def parse_xml(self, xml_file):
        """Parses the HepRep file and returns the root element.
        """
        tree = etree.parse(xml_file)
        root = tree.getroot()
        return root


    def populate_meshes(self, element, component, level=-1, progress_obj=None):
        """Populates the meshes for the given element.

        :param element: The element to populate.
        :type element: xml.etree.ElementTree.Element
        :param component: The component to populate.
        :type component: dict
        :param level: The recursion level.
        :type level: int
        :param progress_obj: The progress object to use for the progress bar.
        :type progress_obj: ProgressBar, optional
        """
        if progress_obj:
            if progress_obj.sync_status(increment=True): return
        
        if element.tag.endswith('instance'):
            # children of instances are attvalues and one primitive
            # loop through the attvalues and set the attributes
            for child in element:
                if progress_obj:
                    if progress_obj.sync_status(increment=True): return

                # get the attributes
                if child.tag.endswith('attvalue'):
                    self.process_attvalue(child, component)

                # now get the point data from the primitive
                elif child.tag.endswith('primitive'):
                    points = []
                    for grandchild in child:
                        if progress_obj:
                            if progress_obj.sync_status(increment=True): return
                            
                        if grandchild.tag.endswith('point'):
                            points.append([float(grandchild.attrib['x']), \
                                           float(grandchild.attrib['y']), \
                                           float(grandchild.attrib['z'])])
                        elif grandchild.tag.endswith('attvalue') and grandchild.attrib['name'].startswith('Radius'):
                            points.append(float(grandchild.attrib['value']))
                    component['points'].append(points)

                # if the child is a type, call the function recursively
                elif child.tag.endswith('type'):
                    self.populate_meshes(child, component, level, progress_obj)

        elif element.tag.endswith('type'):
            # children of types are instances or attvalues
            is_event = False
            name_split = element.attrib['name'].split('_')
            if name_split[-1].isnumeric():
                name = '_'.join(name_split[:-1])
            else:
                name = element.attrib['name']
            if name == 'Event Data':
                self.event_number += 1
            if self.event_number > 0 and (name == 'TransientPolylines' or \
                                          name == 'Hits'):
                # these seem to contain the same information as trajectories
                # so skip them for now
                return
            elif self.event_number > 0 and name == 'Trajectories':
                name = 'Event {} '.format(self.event_number) + name
                is_event = True
                self.event_number += 1
            elif self.event_number > 0 and name == 'Trajectory Step Points':
                is_event = True

            child_component = self.initialize_template(name)
            child_component['is_event'] = is_event

            for child in element:
                if progress_obj:
                    if progress_obj.sync_status(increment=True): return

                if child.tag.endswith('attvalue'):
                    self.process_attvalue(child, child_component)

                elif child.tag.endswith('instance'):
                    self.num_components += 1
                    instance_component = self.initialize_template(name)
                    instance_component['is_event'] = is_event

                    # copy attributes from child_component to instance_component
                    self.copy_parent_attvalues(child_component, instance_component)

                    self.populate_meshes(child, instance_component, level, progress_obj)
                    component['children'].append(instance_component)

            # if no instances were found, add the child_component itself
            if not component['children']:
                component['children'].append(child_component)

        # if not dealing with an instance, try again with each of the children
        else:
            for child in element:
                self.populate_meshes(child, component, level, progress_obj)

        
    def process_attvalue(self, child, component):
        """Processes the attvalue to set the attributes of the component.

        :param child: The attvalue to process.
        :type child: xml.etree.ElementTree.Element
        :param component: The component to set the attributes for.
        :type component: dict
        """
        if child.attrib['name'] == 'DrawAs':
            component['shape'] = child.attrib['value']
        elif child.attrib['name'] == 'LineColor':
            color_str = child.attrib['value']
            color = [float(i)/255. for i in color_str.split(',')]
            component['colors'] = [color]
        elif child.attrib['name'] == 'MarkColor':
            color_str = child.attrib['value']
            color = [float(i)/255. for i in color_str.split(',')]
            component['colors'] = [color]
            component['is_dot'] = True
        elif child.attrib['name'] == 'Visibility':
            component['visible'] = child.attrib['value'] == 'True'


    def copy_parent_attvalues(self, parent, child):
        """Copies the parent attributes to the child.

        :param parent: The parent component.
        :type parent: dict
        :param child: The child component.
        :type child: dict
        """
        child['shape'] = parent['shape']
        child['colors'] = [] if not parent['colors'] else [parent['colors'][0]]
        child['is_dot'] = parent['is_dot']
        child['visible'] = parent['visible']


    def create_meshes(self, components, progress_obj=None):
        """Creates the meshes for the given components.

        :param components: The list of components to create meshes for.
        :type components: list
        """
        for comp in components:
            if progress_obj:
                if progress_obj.sync_status(increment=True): return

            if comp['shape'] == 'Prism':
                comp['mesh_points'] = np.array(comp['points'])
                comp['mesh_inds'] = [[4, 0, 1, 2, 3,\
                                      4, 4, 5, 1, 0,\
                                      4, 7, 4, 0, 3,\
                                      4, 6, 7, 3, 2,\
                                      4, 5, 6, 2, 1,\
                                      4, 7, 6, 5, 4]]
                comp['scalars'] = [comp['colors']*len(comp['points'][0])]
                
            elif comp['shape'] == 'Cylinder':
                points = []
                inds = []
                scalars = []
                if len(comp['points']) == 1:
                    pt, ind = geometry.create_cylinder_mesh(
                        comp['points'][0][2], comp['points'][0][3], \
                        comp['points'][0][0], comp['points'][0][1])
                    points.append(pt)
                    inds.append(ind)
                    scalars.append(comp['colors']*len(pt))
                elif len(comp['points']) == 2:
                    pt, ind = geometry.create_annular_cylinder_mesh(
                        comp['points'][0][2], comp['points'][0][3], \
                        comp['points'][0][0], comp['points'][0][1], \
                        comp['points'][1][0], comp['points'][1][1])
                    points.append(pt)
                    inds.append(ind)
                    scalars.append(comp['colors']*len(pt))
                comp['mesh_points'] = points
                comp['mesh_inds'] = inds
                comp['scalars'] = scalars

            elif comp['shape'] == 'Polygon':
                comp['mesh_points'] = [np.concatenate(comp['points'])]
                unique_points, indices, inverse = np.unique(comp['mesh_points'][0], axis=0,
                                                            return_index=True, return_inverse=True)
                comp['mesh_points'] = [unique_points]

                inds = []
                this_ind = 0
                for point in comp['points']:
                    ind = [len(point)]
                    ind.extend(inverse[this_ind:this_ind + len(point)])
                    this_ind += len(point)
                    inds.append(ind)

                quad_faces = [face for face in inds if len(face) == 5]
                tri_faces = [face for face in inds if len(face) == 4]
                quad_faces = np.unique(np.array(quad_faces), axis=0)
                tri_faces = np.unique(np.array(tri_faces), axis=0)
                comp['mesh_inds'] = [np.concatenate([quad_faces.flatten(), tri_faces.flatten()])]
                comp['scalars'] = [comp['colors'] * len(comp['mesh_points'][0])]

            elif comp['shape'] == 'Point':
                comp['mesh_points'] = np.array(comp['points'])
                comp['mesh_inds'] = [np.array(())]*len(comp['mesh_points'])
                comp['scalars'] = [comp['colors']*len(m) for m in comp['mesh_points']]

            elif comp['shape'] == 'Line':
                comp['mesh_points'] = [np.concatenate(comp['points'])]
                point_inds = np.arange(len(comp['mesh_points'][0]))
                inds = []
                this_ind = 0
                for point in comp['points']:
                    ind = [len(point)]
                    ind.extend(point_inds[this_ind:this_ind + len(point)])
                    this_ind += len(point)
                    inds.append(ind)
                comp['mesh_inds'] = [np.concatenate(inds)]
                comp['scalars'] = [np.concatenate([[color]*len(point) for color, point in \
                                                   zip(comp['colors'], comp['points'])])]

            if len(comp['children']) > 0:
                self.create_meshes(comp['children'], progress_obj)


    def build_mesh_objects(self, components):
        """Draws the meshes for the given components.

        :param components: The list of components to draw meshes for.
        :type components: list
        """
        for comp in components:
            if len(comp['children']) > 0:
                self.build_mesh_objects(comp['children'])
            shape = None
            if comp['shape'] == 'Prism' and len(comp['mesh_points']) > 0 and comp['visible']:
                for i, points in enumerate(comp['mesh_points']):
                    shape = pv.PolyData(points, faces=comp['mesh_inds'][i])

            elif comp['shape'] == 'Cylinder' and comp['visible']:
                for i, points in enumerate(comp['mesh_points']):
                    shape = pv.PolyData(points, faces=comp['mesh_inds'][i])

            elif comp['shape'] == 'Polygon' and comp['visible']:
                for i, points in enumerate(comp['mesh_points']):
                    shape = pv.PolyData(points, faces=comp['mesh_inds'][i])

            elif comp['shape'] == 'Point' and comp['visible']:
                for i, points in enumerate(comp['mesh_points']):
                    shape = pv.PolyData(points)

            elif comp['shape'] == 'Line' and comp['visible']:
                for i, points in enumerate(comp['mesh_points']):
                    shape = pv.PolyData(points, lines=comp['mesh_inds'][i])
            else:
                continue

            if shape is not None:
                shape.point_data.set_scalars(comp['scalars'][i], name='color')

                if shape.n_open_edges > 0:
                    shape, success = self.repair_mesh(shape)

                comp['mesh'] = shape


    def repair_mesh(self, mesh):
        """Attempts to repair the given mesh.
        
        :param mesh: The mesh to repair.
        :type mesh: pv.PolyData
        :return: A tuple containing the repaired mesh and a boolean indicating
            whether the mesh was repaired.
        :rtype: tuple
        """

        cleaned = mesh.clean()
        if cleaned.n_open_edges == 0:
            return cleaned, True
        triangulated = mesh.triangulate()
        if triangulated.n_open_edges == 0:
            return triangulated, True
        else:
            cleaned = triangulated.clean()
            if cleaned.n_open_edges == 0:
                return cleaned, True
        surface = mesh.extract_surface()
        if surface.n_open_edges == 0:
            return surface, True
        else:
            cleaned = surface.clean()
            if cleaned.n_open_edges == 0:
                return cleaned, True
        length = np.linalg.norm(mesh.bounds[1] - mesh.bounds[0])
        filled = mesh.fill_holes(1e-2*length)
        if filled.n_open_edges == 0:
            return filled, True
        else:
            cleaned = filled.clean()
            if cleaned.n_open_edges == 0:
                return cleaned, True

        return mesh, False


    def combine_dicts(self, dicts):
        """Combines the given dictionaries into a single dictionary.

        :param dicts: The list of dictionaries to combine.
        :type dicts: list
        """
        if len(dicts) < 2:
            return dicts[0]
        result = self.initialize_template(dicts[0]['name'])
        result['shape'] = dicts[0]['shape']
        result['visible'] = dicts[0]['visible']
        result['is_dot'] = dicts[0]['is_dot']
        result['is_event'] = dicts[0]['is_event']

        # combine elements in a single dictionary first
        for j in range(len(dicts)):
            if len(dicts[j]['mesh_points']) > 1:
                points, cells, colors = self.combine_mesh_arrays(
                    [dicts[j]['mesh_points'][i] for i in range(len(dicts[j]['mesh_points']))],
                    [dicts[j]['mesh_inds'][i] for i in range(len(dicts[j]['mesh_points']))],
                    [dicts[j]['scalars'][i] for i in range(len(dicts[j]['mesh_points']))])
                dicts[j]['mesh_points'] = [points]
                dicts[j]['mesh_inds'] = [cells]
                dicts[j]['scalars'] = [colors]

        # then combine the dictionaries
        points, cells, colors = self.combine_mesh_arrays(
            [dicts[i]['mesh_points'][0] for i in range(len(dicts))],
            [dicts[i]['mesh_inds'][0] for i in range(len(dicts))],
            [dicts[i]['scalars'][0] for i in range(len(dicts))])
        result['mesh_points'] = [points]
        result['mesh_inds'] = [cells]
        result['scalars'] = [colors]
        children = []
        for d in dicts:
            children.extend(d['children'])
        result['children'] = children
        return result
    
    
    def reduce_components(self, components):
        """Reduces the components by combining duplicate children.

        :param components: The list of components to reduce.
        :type components: list
        """
        for comp in components:
            if len(comp['children']) > 1:
                names = []
                for child in comp['children']:
                    if child['name'] not in names:
                        names.append(child['name'])
                reduced = []
                for name in names:
                    to_combine = [child for child in comp['children'] if child['name'] == name]
                    if len(to_combine) > 1:
                        combined = self.combine_dicts(to_combine)
                        reduced.append(combined)
                    else:
                        reduced.append(to_combine[0])
                comp['children'] = reduced
            self.reduce_components(comp['children'])

        return components



import unittest
from unittest import mock
import os
import tempfile
from geviewer import viewer

class TestGeViewer(unittest.TestCase):

    def setUp(self):
        """Sets up the GeViewer object."""
        self.gev = viewer.GeViewer()

    def test_load_vrml_file(self):
        """Tests the load_file method with a VRML file."""
        self.gev.load_file('tests/sample.wrl', off_screen=True)
        # geometry, step markers, and trajectories should have been loaded
        self.assertEqual(len(self.gev.components[0]['children']), 3)
        
    def test_load_heprep_file(self):
        """Tests the load_file method with a HepRep file."""
        self.gev.load_file('tests/sample.heprep', off_screen=True)
        # geometry and events should have been loaded
        self.assertEqual(len(self.gev.components[0]['children']), 2)

    def test_save_file(self):
        """Tests the save and load methods."""
        with tempfile.TemporaryDirectory() as temp_dir:
            self.gev.off_screen = True
            self.gev.clear_meshes()
            self.gev.load_file('tests/sample.wrl', off_screen=True)
            self.gev.save_session(os.path.join(temp_dir, 'sample.gev'))
            self.assertTrue(os.path.exists(os.path.join(temp_dir, 'sample.gev')))
            self.gev.load_session(os.path.join(temp_dir, 'sample.gev'))
            self.assertEqual(len(self.gev.components[0]['children']), 3)

    def test_clear_meshes(self):
        """Tests the clear_meshes method."""
        self.gev.load_file('tests/sample.wrl', off_screen=True)
        self.gev.clear_meshes()
        self.assertEqual(len(self.gev.plotter.actors), 0)

    def test_find_overlaps(self):
        """Tests the find_overlaps method."""
        self.gev.off_screen = True
        self.gev.clear_meshes()
        self.gev.load_file('tests/sample.heprep', off_screen=True)
        # one overlap between two components
        self.assertEqual(len(self.gev.find_overlaps(tolerance=0.01, n_samples=10000)), 2)

    def test_count_components(self):
        """Tests the count_components method."""
        self.gev.off_screen = True
        self.gev.clear_meshes()
        self.gev.load_file('tests/sample.heprep', off_screen=True)
        self.assertEqual(self.gev.count_components(self.gev.components), 69)

    def test_create_plotter(self):
        """Tests the create_plotter method."""
        self.gev.off_screen = True
        self.gev.clear_meshes()
        self.gev.load_file('tests/sample.heprep', off_screen=True)
        self.gev.create_plotter()
        self.assertEqual(len(self.gev.plotter.actors), 53)

    def test_toggle_parallel_projection(self):
        """Tests the toggle_parallel_projection method."""
        self.gev.off_screen = True
        self.gev.toggle_parallel_projection()
        self.assertEqual(self.gev.plotter.camera.parallel_projection, True)

    def test_toggle_background(self):
        """Tests the toggle_background method."""
        self.gev.load_file('tests/sample.wrl', off_screen=True)
        self.gev.create_plotter()
        self.gev.toggle_background()
        self.assertEqual(self.gev.plotter.background_color, 'white')

    def test_toggle_wireframe(self):
        """Tests the toggle_wireframe method."""
        self.gev.load_file('tests/sample.wrl', off_screen=True)
        self.gev.create_plotter()
        self.gev.toggle_wireframe()
        self.assertEqual(self.gev.wireframe, True)

    def test_toggle_transparent(self):
        """Tests the toggle_transparent method."""
        self.gev.load_file('tests/sample.wrl', off_screen=True)
        self.gev.create_plotter()
        self.gev.toggle_transparent()
        self.assertEqual(next(iter(self.gev.actors.values())).GetProperty().GetOpacity(), 0.3)

if __name__ == '__main__':
    unittest.main()



import numpy as np

def create_cylinder_mesh(p1, p2, r1, r2, num_segments=100):
    """Creates a mesh for a cylinder.

    :param p1: The first endpoint.
    :type p1: tuple
    :param p2: The second endpoint.
    :type p2: tuple
    :param r1: The radius at the first endpoint.
    :type r1: float
    :param r2: The radius at the second endpoint.
    :type r2: float
    :param num_segments: The number of segments to use.
    :type num_segments: int
    """
    # Convert endpoints to numpy arrays
    p1 = np.array(p1)
    p2 = np.array(p2)
    
    # Vector along the cylinder axis
    axis = p2 - p1
    length = np.linalg.norm(axis)
    axis = axis / length
    
    # Arbitrary vector not parallel to axis
    if axis[0] != 0 or axis[1] != 0:
        not_axis = np.array([axis[1], -axis[0], 0])
    else:
        not_axis = np.array([0, axis[2], -axis[1]])
    
    # Orthonormal basis vectors perpendicular to axis
    v = np.cross(axis, not_axis)
    u = np.cross(v, axis)
    u = u / np.linalg.norm(u) * r1
    v = v / np.linalg.norm(v) * r1
    
    # Points on the first end cap
    points = []
    for i in range(num_segments):
        angle = 2 * np.pi * i / num_segments
        point = p1 + np.cos(angle) * u + np.sin(angle) * v
        points.append(point)
    
    # Points on the second end cap
    u = u / r1 * r2
    v = v / r1 * r2
    for i in range(num_segments):
        angle = 2 * np.pi * i / num_segments
        point = p2 + np.cos(angle) * u + np.sin(angle) * v
        points.append(point)
    
    # Indices for the side faces using quadrilaterals
    indices = []
    for i in range(num_segments):
        next_i = (i + 1) % num_segments
        indices.extend([4, i, next_i, next_i + num_segments, i + num_segments])
    
    # Indices for the end caps (triangles)
    center1 = len(points)
    center2 = center1 + 1
    points.append(p1)
    points.append(p2)
    for i in range(num_segments):
        next_i = (i + 1) % num_segments
        indices.extend([3, i, next_i, center1])
        indices.extend([3, i + num_segments, next_i + num_segments, center2])
    
    return points, indices


def create_annular_cylinder_mesh(p1, p2, r1_outer, r2_outer, r1_inner, r2_inner, num_segments=100):
    """Creates a mesh for an annular cylinder.

    :param p1: The first endpoint.
    :type p1: tuple
    :param p2: The second endpoint.
    :type p2: tuple
    :param r1_outer: The outer radius at the first endpoint.
    :type r1_outer: float
    :param r2_outer: The outer radius at the second endpoint.
    :type r2_outer: float
    :param r1_inner: The inner radius at the first endpoint.
    :type r1_inner: float
    :param r2_inner: The inner radius at the second endpoint.
    :type r2_inner: float
    :param num_segments: The number of segments to use.
    :type num_segments: int
    """

    if r1_inner == 0 and r2_inner == 0:
        return create_cylinder_mesh(p1, p2, r1_outer, r2_outer, num_segments)

    # Convert endpoints to numpy arrays
    p1 = np.array(p1)
    p2 = np.array(p2)
    
    # Vector along the cylinder axis
    axis = p2 - p1
    length = np.linalg.norm(axis)
    axis = axis / length
    
    # Arbitrary vector not parallel to axis
    if axis[0] != 0 or axis[1] != 0:
        not_axis = np.array([axis[1], -axis[0], 0])
    else:
        not_axis = np.array([0, axis[2], -axis[1]])
    
    # Orthonormal basis vectors perpendicular to axis
    v = np.cross(axis, not_axis)
    u = np.cross(v, axis)

    def generate_circle_points(center, radius_u, radius_v, num_segments):
        """ Helper function to generate points on a circle. """
        circle_points = []
        for i in range(num_segments):
            angle = 2 * np.pi * i / num_segments
            point = center + np.cos(angle) * radius_u + np.sin(angle) * radius_v
            circle_points.append(point)
        return circle_points

    # Outer points on the first end cap
    u_outer = u / np.linalg.norm(u) * r1_outer
    v_outer = v / np.linalg.norm(v) * r1_outer
    outer_points_1 = generate_circle_points(p1, u_outer, v_outer, num_segments)

    # Outer points on the second end cap
    u_outer = u / np.linalg.norm(u) * r2_outer
    v_outer = v / np.linalg.norm(v) * r2_outer
    outer_points_2 = generate_circle_points(p2, u_outer, v_outer, num_segments)

    # Inner points on the first end cap
    u_inner = u / np.linalg.norm(u) * r1_inner
    v_inner = v / np.linalg.norm(v) * r1_inner
    inner_points_1 = generate_circle_points(p1, u_inner, v_inner, num_segments)

    # Inner points on the second end cap
    u_inner = u / np.linalg.norm(u) * r2_inner
    v_inner = v / np.linalg.norm(v) * r2_inner
    inner_points_2 = generate_circle_points(p2, u_inner, v_inner, num_segments)

    # Combine all points
    points = outer_points_1 + outer_points_2 + inner_points_1 + inner_points_2

    # Indices for the side faces using quadrilaterals
    indices = []
    for i in range(num_segments):
        next_i = (i + 1) % num_segments
        
        # Outer surface
        indices.extend([4, i, next_i, next_i + num_segments, i + num_segments])
        
        # Inner surface
        indices.extend([4, i + 2*num_segments, next_i + 2*num_segments, 
                        next_i + 3*num_segments, i + 3*num_segments])

    # Indices for the end caps (outer to inner ring, using triangles)
    for i in range(num_segments):
        next_i = (i + 1) % num_segments
        
        # First end cap
        indices.extend([4, i, i + 2*num_segments, next_i + 2*num_segments, next_i])
        
        # Second end cap
        indices.extend([4, i + num_segments, i + 3*num_segments, next_i + 3*num_segments, next_i + num_segments])

    return points, indices



import sys
import os
import argparse
from tqdm import tqdm
from pathlib import Path
from geviewer.utils import check_files, print_banner
from geviewer.viewer import GeViewer


class ProgressBar:
    """A progress bar for the converter utility.
    """

    def __init__(self):
        """Initializes the progress bar.
        """
        self.pbar = None
        self.total = 0
        self.interactive = sys.stdout.isatty()


    def reset_progress(self):
        """Resets the progress bar.
        """
        if self.interactive:
            self.pbar = None
            self.total = 0


    def increment_progress(self):
        """Increments the progress bar.
        """
        if self.interactive:
            if self.pbar is None and self.total > 0:
                self.pbar = tqdm(total=self.total)
            if self.pbar is not None:
                if self.pbar.n + 1 < self.total:
                    self.pbar.update(1)


    def set_maximum_value(self, value):
        """Sets the maximum value of the progress bar.
        """
        if self.interactive:
            self.total = value


    def signal_finished(self):
        """Signals that the progress bar is finished.
        """
        if self.interactive:
            if self.pbar is not None:
                self.pbar.n = self.total
                self.pbar = None
                print()


    def print_update(self, text):
        """Prints text sent to the progress bar.
        """
        print(text)


    def sync_status(self, update=None, increment=False):
        """Synchronizes the status of the task with the progress bar.
        """
        if update:
            self.print_update(update)
        if increment:
            self.increment_progress()


def main():
    """Converts the file to .gev format and saves it to the 
    specified destination.
    """
    parser = argparse.ArgumentParser(
        description = 'gev-converter is a command-line utility for GeViewer.',
        epilog = 'For more information, visit https://geviewer.readthedocs.io/en/latest/'
    )
    parser.add_argument('file', help='The file to convert.')
    parser.add_argument('destination', help='The destination to save the converted file to.')
    args = parser.parse_args()

    print_banner()
    file = str(Path(args.file).resolve())
    destination = str(Path(args.destination).resolve())

    if check_files([file]):

        print('Converting {} to {}\n'.format(file, destination))

        dest_dir = os.path.dirname(destination)
        if dest_dir and not os.path.exists(dest_dir):
            os.makedirs(dest_dir)
        if not destination.lower().endswith('.gev'):
            destination += '.gev'
        
        viewer = GeViewer()
        viewer.load_file(file, progress_obj=ProgressBar())
        viewer.save_session(destination)

        print('Success: file saved to {}'.format(destination))
    else:
        print('Error: failed to convert {}'.format(file))


if __name__ == '__main__':
    main()


import sys
import traceback
import webbrowser
import io
import re
from pathlib import Path
import time
import gc
from pyvistaqt import MainWindow

from PyQt6.QtCore import QDateTime, QThread, pyqtSignal, QEventLoop, QTimer

from PyQt6.QtWidgets import (
    QApplication, QVBoxLayout, QHBoxLayout, QPushButton, QWidget,
    QCheckBox, QSplitter, QFileDialog, QLabel, QTextEdit, QLineEdit,
    QColorDialog, QMenu, QToolBar, QSpinBox, QFrame, QScrollArea,
    QMessageBox, QGridLayout, QTabWidget, QGroupBox, QSpacerItem,
    QSizePolicy, QWidgetAction, QGraphicsOpacityEffect, QProgressBar
)
from PyQt6.QtGui import (
    QAction, QFont, QColor, QPalette, QDoubleValidator, QIntValidator,
    QKeySequence, QIcon, QTextCharFormat, QTextCursor, QFontDatabase
)
from PyQt6.QtCore import Qt, pyqtSignal, pyqtSlot, QTimer, QSize, QObject

from geviewer.viewer import GeViewer
import geviewer.utils as utils


class Application(QApplication):
    """A custom application class for the GeViewer application.
    """

    def __init__(self, *args, **kwargs):
        """Initializes the application.
        """
        super().__init__(*args, **kwargs)
        self.window = None

    def notify(self, receiver, event):
        """Handles unhandled exceptions.
        """
        try:
            return super().notify(receiver, event)
        except Exception as e:
            if self.window:
                self.window.global_exception_hook(type(e), e, e.__traceback__)
            return False


class Window(MainWindow):
    """A custom main window class for the GeViewer application.
    """
    #: Signal emitted when the file name changes
    #:
    #: :param str: The new file name
    file_name_changed = pyqtSignal(str)

    #: Signal emitted when the number of events changes
    #:
    #: :param int: The new number of events
    number_of_events = pyqtSignal(int)

    #: Signal emitted when a file is loaded
    file_loaded = pyqtSignal()

    def __init__(self):
        """Initializes the main window.
        """
        super().__init__()

        # initialize some class attributes
        self.default_title = 'GeViewer'
        self.current_file = []
        self.checkbox_mapping = {}
        self.events_list = []
        self.figure_size = [1920, 1440]
        self.worker_running = False
        self.success = True

        # create the main window
        self.setWindowTitle(self.default_title)
        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        main_layout = QHBoxLayout(central_widget)
        splitter = QSplitter(Qt.Horizontal)

        # add panels and menu bar
        self.create_viewer_panel()
        self.create_control_panel()
        self.create_components_panel()
        self.add_menu_bar()

        # create a splitter and add the panels to it
        splitter.addWidget(self.components_panel)
        splitter.addWidget(self.viewer_panel)
        splitter.addWidget(self.control_panel)
        splitter.setChildrenCollapsible(False)
        splitter.setSizes([300, 800, 300])
        main_layout.addWidget(splitter)

        # other setup
        self.number_of_events.connect(self.update_event_total)
        self.file_name_changed.connect(self.update_title)
        self.print_to_console('Welcome to GeViewer!')
        updates = utils.check_for_updates()
        if updates:
            self.print_to_console(updates)
        self.resize_window()
        QApplication.instance().window = self


    def resize_window(self):
        """Resizes the window to fit the screen.
        """
        screen = QApplication.primaryScreen()
        screen_geometry = screen.availableGeometry()

        if int(screen_geometry.width()) < 1920:
            self.showMaximized()
        else:
            self.resize(QSize(int(screen_geometry.width() * 0.8), \
                              int(screen_geometry.height() * 0.8)))
            self.center_on_screen(screen_geometry)


    def center_on_screen(self, screen_geometry):
        """Centers the window on the screen.

        :param screen_geometry: The geometry of the screen.
        :type screen_geometry: QRect
        """
        frame_geometry = self.frameGeometry()
        center_point = screen_geometry.center()
        frame_geometry.moveCenter(center_point)
        self.move(frame_geometry.topLeft())


    def global_exception_hook(self, exctype, value, traceback_obj):
        """Global method to catch unhandled exceptions.
        """
        self.success = False
        error_message = ''.join(traceback.format_exception(exctype, value, traceback_obj))
        self.print_to_console('Error:\n' + error_message)


    def load_initial_files(self, files):
        """Loads initial files after the window has been shown.

        :param files: List of file paths to load
        :type files: list of str
        """

        for file in files:
            self.load_file(file)
            
            # wait for the file to be loaded completely
            loop = QEventLoop()
            self.file_loaded.connect(loop.quit)
            loop.exec()

    ##############################################################
    # Methods for creating the GUI
    ##############################################################

    def create_components_panel(self):
        """Adds the components panel to the main window.
        """
        # create a layout for the components panel
        self.components_panel = QWidget()
        ojbect_layout = QVBoxLayout(self.components_panel)

        # add a heading
        heading = QLabel('Components')
        tab_height = self.tab_widget.tabBar().height()
        heading.setFixedHeight(tab_height)
        heading.setAlignment(Qt.AlignVCenter)
        heading.setStyleSheet("padding-left: 5px;")
        ojbect_layout.addWidget(heading)

        # create the scroll area for the checkboxes
        components_area = QScrollArea()
        components_area.setWidgetResizable(True)

        # add temporary instructions
        instructions = 'Click "File > Open File..." to add components'
        self.load_instructions = QLabel(instructions)
        load_instructions_font = QFont()
        load_instructions_font.setPointSize(13)
        load_instructions_font.setItalic(True)
        opacity_effect = QGraphicsOpacityEffect()
        opacity_effect.setOpacity(0.5)
        self.load_instructions.setGraphicsEffect(opacity_effect)
        self.load_instructions.setFont(load_instructions_font)
        self.load_instructions.setWordWrap(True)

        # create a widget to hold the checkboxes
        checkboxes_widget = QWidget()
        self.checkboxes_layout = QVBoxLayout(checkboxes_widget)
        self.checkboxes_layout.setAlignment(Qt.AlignTop)
        self.checkboxes_layout.addWidget(self.load_instructions)
        components_area.setWidget(checkboxes_widget)
        ojbect_layout.addWidget(components_area)


    def create_viewer_panel(self):
        """Adds the viewer panel to the main window.
        """
        # create a layout for the viewer panel
        self.viewer_panel = QWidget()
        self.viewer_layout = QVBoxLayout(self.viewer_panel)

        # create the viewer
        self.viewer = GeViewer(self.viewer_panel)
        self.plotter = self.viewer.plotter
        self.add_key_events()

        # add the toolbar
        self.add_toolbar()

        # add the plotter to the viewer layout
        self.viewer_layout.addWidget(self.plotter.interactor)


    def create_control_panel(self):
        """Adds the control panel to the main window.
        """
        # create a layout for the control panel
        self.control_panel = QWidget()
        control_layout = QVBoxLayout(self.control_panel)

        # add a heading for the console
        control_panel_heading = QLabel('Control Panel')
        control_panel_heading_font = QFont()
        control_panel_heading_font.setPointSize(14)
        control_panel_heading_font.setBold(True)
        control_panel_heading.setFont(control_panel_heading_font)

        # create a tab widget
        self.tab_widget = QTabWidget()
        self.tab_widget.setMinimumWidth(250)
        self.tab_widget.setMinimumHeight(300)
        self.tab_widget.setMaximumHeight(350)

        # create tabs
        options_tab = QWidget()
        tools_tab = QWidget()
        self.options_layout = QVBoxLayout(options_tab)
        self.tools_layout = QVBoxLayout(tools_tab)
        self.tab_widget.addTab(options_tab, 'Options')
        self.tab_widget.addTab(tools_tab, 'Tools')
        control_layout.addWidget(self.tab_widget)

        # add components to tabs
        self.add_options_tab()
        self.add_tools_tab()

        # create the console
        console_layout = QVBoxLayout()
        self.console = QTextEdit()
        self.console.setReadOnly(True)
        font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        font.setPointSize(13)
        self.console.setFont(font)
        sys.stdout = ConsoleRedirect(self.console)
        sys.stderr = ConsoleRedirect(self.console)
        console_layout.addWidget(self.console)
        control_layout.addLayout(console_layout)

        # Add the progress bar
        self.progress_bar = ProgressBar()
        control_layout.addWidget(self.progress_bar)


    def add_options_tab(self):
        """Adds camera and figure options to the first tab."""
        grid_layout = QGridLayout()

        # camera options section
        heading = QLabel('Camera Options')
        heading_font = QFont()
        heading_font.setPointSize(14)
        heading_font.setBold(True)
        heading.setFont(heading_font)
        grid_layout.addWidget(heading, 0, 0, 1, 2)

        self.camera_position_label = QLabel('Position:')
        self.camera_position_text = QLineEdit()
        self.camera_position_text.setReadOnly(False)
        self.camera_position_text.editingFinished.connect(self.handle_camera_position_change)
        self.camera_position_text.setToolTip('Enter the camera coordinates as three comma-separated numbers')
        grid_layout.addWidget(self.camera_position_label, 1, 0)
        grid_layout.addWidget(self.camera_position_text, 1, 1, 1, 1)

        self.camera_focal_label = QLabel('Focal point:')
        self.camera_focal_text = QLineEdit()
        self.camera_focal_text.setReadOnly(False)
        self.camera_focal_text.editingFinished.connect(self.handle_camera_focal_change)
        self.camera_focal_text.setToolTip('Enter the focal point coordinates as three comma-separated numbers')
        grid_layout.addWidget(self.camera_focal_label, 2, 0)
        grid_layout.addWidget(self.camera_focal_text, 2, 1, 1, 1)

        self.camera_up_label = QLabel('Up vector:')
        self.camera_up_text = QLineEdit()
        self.camera_up_text.setReadOnly(False)
        self.camera_up_text.editingFinished.connect(self.handle_camera_up_change)
        self.camera_up_text.setToolTip('Enter the up vector components as three comma-separated numbers')
        grid_layout.addWidget(self.camera_up_label, 3, 0)
        grid_layout.addWidget(self.camera_up_text, 3, 1, 1, 1)

        # event viewer section
        event_heading = QLabel('Event Viewer')
        event_heading_font = QFont()
        event_heading_font.setPointSize(14)
        event_heading_font.setBold(True)
        event_heading.setFont(event_heading_font)
        grid_layout.addWidget(event_heading, 4, 0, 1, 2)

        self.event_selection_label = QLabel('Show event:')
        self.event_selection_box = QSpinBox()
        self.event_selection_box.setRange(1, 1)
        self.event_selection_box.setValue(1)
        self.event_selection_box.setWrapping(True)
        self.event_selection_box.valueChanged.connect(lambda: self.show_single_event(self.event_selection_box.value()))
        grid_layout.addWidget(self.event_selection_label, 5, 0)
        grid_layout.addWidget(self.event_selection_box, 5, 1, 1, 1)

        # figure options section
        figure_heading = QLabel('Figure Options')
        figure_heading_font = QFont()
        figure_heading_font.setPointSize(14)
        figure_heading_font.setBold(True)
        figure_heading.setFont(figure_heading_font)
        grid_layout.addWidget(figure_heading, 6, 0, 1, 2)

        self.figure_size_label = QLabel('Figure size:')
        self.figure_size_text = QLineEdit()
        self.figure_size_text.setText('{}, {}'.format(*self.figure_size))
        self.figure_size_text.setReadOnly(False)
        self.figure_size_text.editingFinished.connect(self.handle_figure_size_change)
        self.figure_size_text.setToolTip('Enter the figure width and height in pixels as two comma-separated numbers')
        grid_layout.addWidget(self.figure_size_label, 7, 0)
        grid_layout.addWidget(self.figure_size_text, 7, 1, 1, 1)

        self.save_button = QPushButton('Export Figure', self)
        self.save_button.clicked.connect(self.export_figure_dialog)
        grid_layout.addWidget(self.save_button, 8, 0, 1, 2)

        self.options_layout.addLayout(grid_layout)


    def add_tools_tab(self):
        """Adds view and geometry options to the second tab."""
        grid_layout = QGridLayout()

        # overlap inspector section
        overlap_heading = QLabel('Overlap Inspector')
        overlap_heading_font = QFont()
        overlap_heading_font.setPointSize(14)
        overlap_heading_font.setBold(True)
        overlap_heading.setFont(overlap_heading_font)
        grid_layout.addWidget(overlap_heading, 0, 0, 1, 2)

        tolerance_label = QLabel('Tolerance:')
        tolerance_box = QLineEdit()
        double_validator = QDoubleValidator(1e-6, 1, 6)
        double_validator.setNotation(QDoubleValidator.Notation.StandardNotation)
        tolerance_box.setValidator(double_validator)
        tolerance_box.setText('0.001')
        tolerance_box.setToolTip('Enter the tolerance for overlap detection')
        grid_layout.addWidget(tolerance_label, 1, 0)
        grid_layout.addWidget(tolerance_box, 1, 1, 1, 1)

        samples_label = QLabel('Number of points:')
        samples_box = QLineEdit(self)
        int_validator = QIntValidator(1000, 10000000)
        samples_box.setValidator(int_validator)
        samples_box.setText('10000')
        samples_box.setToolTip('Enter the number of sample points for overlap detection')
        grid_layout.addWidget(samples_label, 2, 0)
        grid_layout.addWidget(samples_box, 2, 1, 1, 1)

        check_button = QPushButton('Find Overlaps')
        check_button.clicked.connect(lambda: self.check_geometry(tolerance_box.text(), samples_box.text()))
        clear_overlaps_button = QPushButton('Clear')
        clear_overlaps_button.clicked.connect(self.clear_overlaps)
        grid_layout.addWidget(check_button, 3, 0)
        grid_layout.addWidget(clear_overlaps_button, 3, 1, 1, 1)

        # measurement tool section
        measurement_heading = QLabel('Measurement Tool')
        measurement_heading_font = QFont()
        measurement_heading_font.setPointSize(14)
        measurement_heading_font.setBold(True)
        measurement_heading.setFont(measurement_heading_font)
        grid_layout.addWidget(measurement_heading, 4, 0, 1, 2)

        measure_text = QLabel('Measurement 1:')
        self.measurement_box = QLineEdit()
        self.measurement_box.setReadOnly(True)
        grid_layout.addWidget(measure_text, 5, 0)
        grid_layout.addWidget(self.measurement_box, 5, 1, 1, 1)

        measure_text_2 = QLabel('Measurement 2:')
        self.measurement_box_2 = QLineEdit()
        self.measurement_box_2.setReadOnly(True)
        grid_layout.addWidget(measure_text_2, 6, 0)
        grid_layout.addWidget(self.measurement_box_2, 6, 1, 1, 1)

        measure_text_3 = QLabel('Measurement 3:')
        self.measurement_box_3 = QLineEdit()
        self.measurement_box_3.setReadOnly(True)
        grid_layout.addWidget(measure_text_3, 7, 0)
        grid_layout.addWidget(self.measurement_box_3, 7, 1, 1, 1)

        measure_button = QPushButton('Add Measurement')
        measure_button.setMinimumWidth(130)
        measure_button.clicked.connect(self.measure_distance)
        clear_measurement_button = QPushButton('Clear')
        clear_measurement_button.clicked.connect(self.clear_measurement)
        grid_layout.addWidget(measure_button, 8, 0)
        grid_layout.addWidget(clear_measurement_button, 8, 1)

        # timer to update the camera settings when the view is changed
        self.update_timer = QTimer()
        self.update_timer.setInterval(200)
        self.update_timer.setSingleShot(True)
        self.update_timer.timeout.connect(self.update_view_params)
        self.last_camera_position = None
        self.last_figure_size = None
        self.monitor_camera_position()

        self.tools_layout.addLayout(grid_layout)


    def add_toolbar(self):
        """Adds the toolbar to the main window.

        This method creates a QToolBar and populates it with various actions
        for controlling the viewer's display options and camera views. The
        toolbar includes actions for toggling wireframe mode, transparency,
        parallel projection, and setting different standard views (isometric,
        top, bottom, front, back).
        """
        self.toolbar = QToolBar('Main Toolbar')
        self.toolbar.setMovable(True)

        # text and button size
        toolbar_font = QFont()
        toolbar_font.setPointSize(13)
        action_width = 80
        self.toolbar.setFont(toolbar_font)

        # actions that change when clicked
        self.wireframe_action = QAction('Wireframe', self)
        self.wireframe_action.triggered.connect(self.toggle_wireframe)

        self.transparent_action = QAction('Transparent', self)
        self.transparent_action.triggered.connect(self.toggle_transparent)

        self.parallel_action = QAction('Parallel', self)
        self.parallel_action.triggered.connect(self.toggle_parallel)

        self.toolbar.addAction(self.wireframe_action)
        tool_button = self.toolbar.widgetForAction(self.wireframe_action)
        tool_button.setFixedWidth(action_width)
        self.toolbar.addAction(self.transparent_action)
        tool_button = self.toolbar.widgetForAction(self.transparent_action)
        tool_button.setFixedWidth(action_width)
        self.toolbar.addAction(self.parallel_action)
        tool_button = self.toolbar.widgetForAction(self.parallel_action)
        tool_button.setFixedWidth(action_width)

        # view actions that don't change when clicked
        view_actions = [
            ('Isometric', lambda: (self.print_to_console('Switching to isometric view.'), \
                                   self.plotter.view_isometric())),
            ('Top', lambda: (self.print_to_console('Switching to top view.'), \
                             self.plotter.view_xy())),
            ('Bottom', lambda: (self.print_to_console('Switching to bottom view.'), \
                                self.plotter.view_xy(negative=True))),
            ('Front', lambda: (self.print_to_console('Switching to front view.'), \
                               self.plotter.view_yz())),
            ('Back', lambda: (self.print_to_console('Switching to back view.'), \
                              self.plotter.view_yz(negative=True))),
            ('Left', lambda: (self.print_to_console('Switching to left view.'), \
                              self.plotter.view_xz())),
            ('Right', lambda: (self.print_to_console('Switching to right view.'), \
                               self.plotter.view_xz(negative=True)))
        ]

        for text, callback in view_actions:
            action = QAction(text, self)
            action.triggered.connect(callback)
            self.toolbar.addAction(action)
            tool_button = self.toolbar.widgetForAction(action)
            tool_button.setFixedWidth(60)

        self.viewer_layout.addWidget(self.toolbar)


    def add_menu_bar(self):
        """Adds the menu bar to the main window.

        This method creates and configures the main menu bar for the application.
        It adds the following menus:
        - File: Contains actions for opening files, saving, and closing the window.
        - Edit: Contains actions for clearing the console, copying console content, and clearing meshes.
        - View: Contains actions for toggling visibility of various panels and visual elements.
        - Window: Contains actions for closing the window.
        - Help: Contains actions for displaying the license information.

        The menu bar provides easy access to key functionality and settings of the application.
        """ 

        # create the menu bar
        menubar = self.menuBar()

        # create the file menu
        file_menu = menubar.addMenu('File')
        open_action = file_menu.addAction('Open File...')
        open_action.triggered.connect(self.open_file_dialog)
        open_action.setShortcut(QKeySequence.Open)
        save_action = file_menu.addAction('Save As...')
        save_action.setShortcut(QKeySequence.Save)
        save_action.triggered.connect(self.save_file_dialog)
        close_window_action = file_menu.addAction('Close Window')
        close_window_action.triggered.connect(self.close)
        close_window_action.setShortcut(QKeySequence.Close)

        # create the edit menu
        edit_menu = menubar.addMenu('Edit')
        copy_console_action = edit_menu.addAction('Copy Console')
        copy_console_action.triggered.connect(self.console.selectAll)
        copy_console_action.triggered.connect(self.console.copy)
        clear_console_action = edit_menu.addAction('Clear Console')
        clear_console_action.triggered.connect(self.console.clear)
        clear_action = edit_menu.addAction('Clear Viewer')
        clear_action.triggered.connect(self.clear_viewer)
        abort_action = edit_menu.addAction('Abort Process')
        abort_action.triggered.connect(self.abort_process)

        # create the view menu
        view_menu = menubar.addMenu('View')
        show_components_action = QAction('Show Components Panel', self, checkable=True)
        show_components_action.setChecked(True)
        show_components_action.triggered.connect(self.toggle_components_panel)
        view_menu.addAction(show_components_action)

        # create the show controls action
        self.show_controls_action = QAction('Show Control Panel', self, checkable=True)
        self.show_controls_action.setChecked(True)
        self.show_controls_action.triggered.connect(self.toggle_control_panel)
        view_menu.addAction(self.show_controls_action)

        # create the show background action
        show_background_action = QAction('Show Background', self, checkable=True)
        show_background_action.setChecked(True)
        show_background_action.triggered.connect(self.toggle_background)
        view_menu.addAction(show_background_action)

        # create the show gradient action
        gradient_action = QAction('Show Gradient', self, checkable=True)
        gradient_action.setChecked(True)
        gradient_action.triggered.connect(self.toggle_gradient)
        view_menu.addAction(gradient_action)

        # create the depth peeling action
        self.depth_peeling_action = QAction('Use Depth Peeling', self, checkable=True)
        self.depth_peeling_action.setChecked(False)
        self.depth_peeling_action.triggered.connect(self.toggle_depth_peeling)
        view_menu.addAction(self.depth_peeling_action)

        # create the background color menu
        color_menu = QMenu('Background Colors', self)
        color_action_1 = QAction('Set Primary Color...', self)
        color_action_2 = QAction('Set Secondary Color...', self)
        color_action_1.triggered.connect(lambda: self.open_color_picker(0))
        color_action_2.triggered.connect(lambda: self.open_color_picker(1))
        color_menu.addAction(color_action_1)
        color_menu.addAction(color_action_2)
        view_menu.addMenu(color_menu)

        # create the reset background action
        reset_background_action = QAction('Reset Background', self)
        reset_background_action.triggered.connect(lambda: self.open_color_picker(2))
        view_menu.addAction(reset_background_action)

        # create the window menu
        window_menu = menubar.addMenu('Window')
        maximize_action = window_menu.addAction('Maximize')
        maximize_action.triggered.connect(self.showMaximized)
        minimize_action = window_menu.addAction('Minimize')
        minimize_action.triggered.connect(self.showMinimized)
        restore_action = window_menu.addAction('Restore')
        restore_action.triggered.connect(self.showNormal)

        # create the help menu
        help_menu = menubar.addMenu('Help')
        license_action = help_menu.addAction('License')
        license_action.triggered.connect(self.show_license)
        update_action = help_menu.addAction('Check for Updates')
        update_action.triggered.connect(self.check_for_updates)
        documentation_action = help_menu.addAction('Documentation')
        documentation_action.triggered.connect(self.show_documentation)

    ##############################################################
    # Signals and slots
    ##############################################################
    
    @pyqtSlot(str)
    def update_title(self, filename):
        """Updates the title text to reflect the current file being viewed.

        This method updates the text of the window title to indicate the current file
        being viewed. If multiple files are being viewed, it will also show
        the number of additional files being viewed.

        :param filename: The name of the file being viewed.
        :type filename: str
        """
        if filename:
            self.current_file.append(filename)
            title = self.default_title + ' - ' + str(Path(self.current_file[0]).resolve()) \
                    + ['',' + {} more'.format(len(self.current_file) - 1)][len(self.current_file) > 1]
        else:
            title = self.default_title
        self.setWindowTitle(title)


    @pyqtSlot(int)
    def update_event_total(self, total):
        """Updates the event selection box to reflect the total number of events.

        This method updates the range of the event selection box to reflect the
        total number of events in the current file.

        :param total: The total number of events in the current file.
        :type total: int
        """
        self.event_selection_box.setRange(1, max(1, total))

    ##############################################################
    # Methods for the components panel
    ##############################################################

    def generate_checkboxes(self, components, level):
        """Generates checkboxes for the components in the current file.

        This method generates checkboxes for the components in the current file.
        It also sets up the connections for toggling visibility and updating
        the event selection box.

        :param components: The components to generate checkboxes for.
        :type components: list
        :param level: The level of the components to generate checkboxes for.
        :type level: int
        """
        self.checkboxes_layout.removeWidget(self.load_instructions)
        self.load_instructions.setParent(None)
        self.load_instructions.hide()
        for comp in components:
            if comp['id'] not in self.checkbox_mapping:
                checkbox = QCheckBox(comp['name'])
                checkbox.setCheckState(Qt.CheckState.Checked)
                checkbox.stateChanged.connect(lambda state, comp=comp: self.toggle_visibility(state, comp))
                self.checkboxes_layout.addWidget(checkbox)
                checkbox.setStyleSheet('padding-left: {}px'.format(20 * level))
                self.checkbox_mapping[comp['id']] = checkbox
                if comp['is_event'] and 'Event' in comp['name']:
                    self.events_list.append(comp['id'])
                if 'children' in comp and comp['children']:
                    self.generate_checkboxes(comp['children'], level + 1)
        self.number_of_events.emit(len(self.events_list))


    def toggle_visibility(self, state, comp):
        """Toggles the visibility of a component in the viewer.

        This method toggles the visibility of a component in the viewer by
        checking the corresponding checkbox and updating the visibility of the
        associated actors.

        :param state: The state of the checkbox to toggle.
        :type state: int
        :param comp: The component to toggle visibility for.
        :type comp: dict
        """
        visibility = state > 0
        if comp['has_actor']:
            self.viewer.actors[comp['id']].SetVisibility(visibility)
        if 'children' in comp and comp['children']:
            for child in comp['children']:
                self.set_visibility_recursive(child, visibility)
    

    def set_visibility_recursive(self, comp, visibility):
        """Sets the visibility of a component and all its children recursively.

        This method sets the visibility of a component and all its children
        recursively by checking the corresponding checkbox and updating the
        visibility of the associated actors.

        :param comp: The component to set visibility for.
        :type comp: dict
        :param visibility: The visibility to set for the component.
        :type visibility: bool
        """
        state = Qt.CheckState.Checked if visibility else Qt.CheckState.Unchecked
        if comp['has_actor']:
            self.viewer.actors[comp['id']].SetVisibility(visibility)
        self.checkbox_mapping[comp['id']].setCheckState(state)
        if 'children' in comp and comp['children']:
            for child in comp['children']:
                self.set_visibility_recursive(child, visibility)

    ##############################################################
    # Methods for updating the viewer
    ##############################################################

    def show_single_event(self, event_index):
        """Shows a single event in the viewer.

        This method shows a single event in the viewer by checking the
        corresponding checkbox and updating the visibility of the associated
        actors.

        :param event_index: The index of the event to show.
        :type event_index: int
        """
        self.print_to_console('Showing event {}.'.format(event_index))
        for i, event in enumerate(self.events_list):
            if i == event_index - 1:
                self.checkbox_mapping[event].setCheckState(Qt.CheckState.Checked)
            else:
                self.checkbox_mapping[event].setCheckState(Qt.CheckState.Unchecked)


    def toggle_parallel(self):
        """Toggles the parallel projection.

        This method toggles the parallel projection by setting the parallel
        attribute of the viewer to the opposite of its current value and
        updating the parallel button.
        """
        self.print_to_console('Turning parallel projection ' + ['on.','off.'][self.viewer.parallel])
        self.viewer.toggle_parallel_projection()
        self.parallel_action.setText('Perspective' if self.viewer.parallel else 'Parallel')


    def toggle_wireframe(self):
        """Toggles the wireframe.

        This method toggles the wireframe by setting the wireframe attribute
        of the viewer to the opposite of its current value and updating the
        wireframe button.
        """
        self.print_to_console('Switching to ' + ['wireframe', 'solid'][self.viewer.wireframe] + ' mode.')
        self.viewer.toggle_wireframe()
        self.update_wireframe_action()


    def toggle_transparent(self):
        """Toggles the transparency.

        This method toggles the transparency by setting the transparent attribute
        of the viewer to the opposite of its current value and updating the
        transparency button.
        """
        self.print_to_console('Switching to ' + ['transparent', 'opaque'][self.viewer.transparent] + ' mode.')
        self.viewer.toggle_transparent()
        self.transparent_action.setText('Opaque' if self.viewer.transparent else 'Transparent')


    def update_wireframe_action(self):
        """Updates the wireframe button.

        This method updates the wireframe button by setting the text to 'Solid'
        if the wireframe is enabled, otherwise 'Wireframe'.
        """
        self.wireframe_action.setText('Solid' if self.viewer.wireframe else 'Wireframe')


    def add_key_events(self):
        """Adds key events to the plotter.

        This is the simplest way to have the buttons synchronize
        with whatever key inputs the user provides.
        """
        self.plotter.add_key_event('w', lambda: self.synchronize_toolbar(True))
        self.plotter.add_key_event('s', lambda: self.synchronize_toolbar(False))


    def synchronize_toolbar(self, wireframe=True):
        """Synchronizes the toolbar.

        This method synchronizes the toolbar by setting the wireframe attribute
        of the viewer to the opposite of its current value and updating the
        wireframe button.

        :param wireframe: The wireframe state to synchronize.
        :type wireframe: bool
        """
        self.viewer.wireframe = wireframe
        self.update_wireframe_action()

    ##############################################################
    # Methods for the options tab
    ##############################################################

    def set_camera_position(self, position):
        """Sets the camera position in the viewer.

        This method sets the camera position in the viewer by updating the
        camera's position attribute and triggering an update of the plotter.

        :param position: The new camera position.
        :type position: str
        """
        self.print_to_console('Setting camera position to ' + position + '.')
        position = [float(x) for x in position.split(',')]
        self.plotter.camera.position = position
        self.plotter.update()


    def set_camera_focal(self, focal_point):
        """Sets the camera focal point in the viewer.

        This method sets the camera focal point in the viewer by updating the
        camera's focal point attribute and triggering an update of the plotter.

        :param focal_point: The new camera focal point.
        :type focal_point: str
        """
        self.print_to_console('Setting camera focal point to ' + focal_point + '.')
        focal_point = [float(x) for x in focal_point.split(',')]
        self.plotter.camera.focal_point = focal_point
        self.plotter.update()


    def set_camera_up(self, up_vector):
        """Sets the camera up vector in the viewer.

        This method sets the camera up vector in the viewer by updating the
        camera's up vector attribute and triggering an update of the plotter.

        :param up_vector: The new camera up vector.
        :type up_vector: str
        """
        self.print_to_console('Setting camera up vector to ' + up_vector + '.')
        up_vector = [float(x) for x in up_vector.split(',')]
        self.plotter.camera.up = up_vector
        self.plotter.update()


    def set_figure_size(self, figure_size):
        """Sets the figure size in the viewer.

        This method sets the figure size in the viewer by updating the
        figure size attribute and triggering an update of the plotter.

        :param figure_size: The new figure size.
        :type figure_size: str
        """
        self.print_to_console('Setting figure size to ' + figure_size + '.')
        figure_size = [int(x) for x in figure_size.split(',')]
        self.figure_size = figure_size


    def handle_camera_position_change(self):
        """Handles the change in camera position.

        This method handles the change in camera position by validating the
        new position and setting the camera position in the viewer.
        """
        new_position = self.camera_position_text.text()
        if self.validate_camera_position(new_position):
            self.set_camera_position(new_position)
            self.clear_position_error_state()
        else:
            self.set_position_error_state()


    def handle_camera_focal_change(self):
        """Handles the change in camera focal point.

        This method handles the change in camera focal point by validating the
        new focal point and setting the camera focal point in the viewer.
        """
        new_focal = self.camera_focal_text.text()
        if self.validate_camera_focal(new_focal):
            self.set_camera_focal(new_focal)
            self.clear_focal_error_state()
        else:
            self.set_focal_error_state()


    def handle_camera_up_change(self):
        """Handles the change in camera up vector.

        This method handles the change in camera up vector by validating the
        new up vector and setting the camera up vector in the viewer.
        """
        new_up = self.camera_up_text.text()
        if self.validate_camera_up(new_up):
            self.set_camera_up(new_up)
            self.clear_up_error_state()
        else:
            self.set_up_error_state()


    def handle_figure_size_change(self):
        """Handles the change in figure size.

        This method handles the change in figure size by validating the
        new size and setting the figure size in the viewer.
        """
        new_size = self.figure_size_text.text()
        if self.validate_figure_size(new_size):
            self.set_figure_size(new_size)
            self.clear_window_error_state()
        else:
            self.set_window_error_state()


    def validate_camera_position(self, position):
        """Validates the camera position.

        This method validates the camera position by checking if it is a
        comma-separated list of three floats. If not, it will print an error
        message and return False.

        :param position: The camera position to validate.
        :type position: str
        :return: True if the camera position is valid, False otherwise.
        :rtype: bool
        """
        try:
            position = [float(x) for x in position.split(',')]
            if len(position) != 3:
                raise ValueError('Invalid camera position')
            return True
        except ValueError:
            self.print_to_console('Error: invalid camera position. Please enter three comma-separated floats.')
            return False
        

    def validate_camera_focal(self, focal_point):
        """Validates the camera focal point.

        This method validates the camera focal point by checking if it is a
        comma-separated list of three floats. If not, it will print an error
        message and return False.

        :param focal_point: The camera focal point to validate.
        :type focal_point: str
        :return: True if the camera focal point is valid, False otherwise.
        :rtype: bool
        """
        try:
            focal_point = [float(x) for x in focal_point.split(',')]
            if len(focal_point) != 3:
                raise ValueError('Invalid camera focal point')
            return True
        except ValueError:
            self.print_to_console('Error: invalid camera focal point. Please enter three comma-separated floats.')
            return False
        
        
    def validate_camera_up(self, up_vector):
        """Validates the camera up vector.

        This method validates the camera up vector by checking if it is a
        comma-separated list of three floats. If not, it will print an error
        message and return False.

        :param up_vector: The camera up vector to validate.
        :type up_vector: str
        :return: True if the camera up vector is valid, False otherwise.
        :rtype: bool
        """
        try:
            up_vector = [float(x) for x in up_vector.split(',')]
            if len(up_vector) != 3:
                raise ValueError('Invalid camera up vector')
            return True
        except ValueError:
            self.print_to_console('Error: invalid camera up vector. Please enter three comma-separated floats.')
            return False
        
        
    def validate_figure_size(self, figure_size):
        """Validates the figure size.

        This method validates the figure size by checking if it is a
        comma-separated list of two integers. If not, it will print an error
        message and return False.

        :param figure_size: The figure size to validate.
        :type figure_size: str
        :return: True if the figure size is valid, False otherwise.
        :rtype: bool
        """
        try:
            figure_size = [int(x) for x in figure_size.split(',')]
            if len(figure_size) != 2:
                raise ValueError('Invalid figure size')
            return True
        except ValueError:
            self.print_to_console('Error: invalid figure size. Please enter two comma-separated integers.')
            return False


    def clear_position_error_state(self):
        """Clears the position error state.

        This method clears the position error state by setting the palette
        of the camera position text to the default palette.
        """
        self.camera_position_text.setPalette(QApplication.palette())


    def clear_focal_error_state(self):
        """Clears the focal error state.

        This method clears the focal error state by setting the palette
        of the camera focal text to the default palette.
        """
        self.camera_focal_text.setPalette(QApplication.palette())


    def clear_up_error_state(self):
        """Clears the up error state.

        This method clears the up error state by setting the palette
        of the camera up text to the default palette.
        """
        self.camera_up_text.setPalette(QApplication.palette())


    def clear_window_error_state(self):
        """Clears the window error state.

        This method clears the window error state by setting the palette
        of the figure size text to the default palette.
        """
        self.figure_size_text.setPalette(QApplication.palette())


    def set_position_error_state(self):
        """Sets the position error state.

        This method sets the position error state by setting the palette
        of the camera position text to a light red color.
        """
        palette = self.camera_position_text.palette()
        palette.setColor(QPalette.Base, QColor(255, 192, 192))
        self.camera_position_text.setPalette(palette)


    def set_focal_error_state(self):
        """Sets the focal error state.

        This method sets the focal error state by setting the palette
        of the camera focal text to a light red color.
        """
        palette = self.camera_focal_text.palette()
        palette.setColor(QPalette.Base, QColor(255, 192, 192))
        self.camera_focal_text.setPalette(palette)


    def set_up_error_state(self):
        """Sets the up error state.

        This method sets the up error state by setting the palette
        of the camera up text to a light red color.
        """
        palette = self.camera_up_text.palette()
        palette.setColor(QPalette.Base, QColor(255, 192, 192))
        self.camera_up_text.setPalette(palette)


    def set_window_error_state(self):
        """Sets the window error state.

        This method sets the window error state by setting the palette
        of the figure size text to a light red color.
        """
        palette = self.figure_size_text.palette()
        palette.setColor(QPalette.Base, QColor(255, 192, 192))
        self.figure_size_text.setPalette(palette)


    def update_menu_action(self, visible):
        """Updates the menu action.

        This method updates the menu action by setting the check state
        to checked if visible, otherwise unchecked.

        :param visible: The visibility of the menu action.
        :type visible: bool
        """
        state = Qt.CheckState.Checked if visible else Qt.CheckState.Unchecked
        self.show_controls_action.setCheckState(state)


    def monitor_camera_position(self):
        """Monitors the camera position.

        This method monitors the camera position by checking if it has changed
        from the last position and updating the timer if it has.
        """
        camera_position = self.plotter.camera_position
        if camera_position != self.last_camera_position:
            self.last_camera_position = camera_position
            self.update_timer.start()
        QTimer.singleShot(100, self.monitor_camera_position)


    def update_view_params(self):
        """Updates the view parameters.

        This method updates the view parameters by setting the camera position,
        focal point, and up vector in the viewer.
        """
        camera_pos = self.plotter.camera_position
        self.camera_position_text.setText('{:.3f}, {:.3f}, {:.3f}'.format(*camera_pos[0]))
        self.camera_focal_text.setText('{:.3f}, {:.3f}, {:.3f}'.format(*camera_pos[1]))
        self.camera_up_text.setText('{:.3f}, {:.3f}, {:.3f}'.format(*camera_pos[2]))
        self.clear_position_error_state()
        self.clear_focal_error_state()
        self.clear_up_error_state()


    def export_figure_dialog(self):
        """Saves the file dialog.

        This method saves the file dialog by getting the save file name
        and emitting the save figure event.
        """
        options = QFileDialog.Options()
        file_types = 'Supported File Types (*.png *.jpeg *.jpg *.bmp *.tif *.tiff *.svg *.eps *.ps *.pdf *.tex);; '\
                     + 'PNG (*.png);;JPEG (*.jpeg);;JPG (*.jpg);;BMP (*.bmp);;TIF (*.tif);;TIFF (*.tiff);;'\
                     + 'SVG (*.svg);;EPS (*.eps);;PS (*.ps);;PDF (*.pdf);;TEX (*.tex)'
        file_name, _ = QFileDialog.getSaveFileName(self, 'Save Figure', '', file_types, options=options)
        
        if file_name:
            self.export_figure(file_name, *self.figure_size)

                
    def export_figure(self, file_name, width, height):
        """Saves the figure.

        This method saves the figure by creating an off-screen plotter with the
        desired size and copying the mesh and camera settings to the off-screen
        plotter.

        :param file_name: The file name to save the figure to.
        :type file_name: str
        :param width: The width of the figure.
        :type width: int
        :param height: The height of the figure.
        :type height: int
        """
        import pyvista as pv

        # create an off-screen plotter with the given size
        off_screen_plotter = pv.Plotter(off_screen=True, window_size=[width, height])

        if self.viewer.bkg_on:
            if self.viewer.gradient:
                off_screen_plotter.set_background(*self.viewer.bkg_colors)
            else:
                off_screen_plotter.set_background(self.viewer.bkg_colors[0])
        else:
            off_screen_plotter.set_background('white')

        if self.depth_peeling_action.isChecked():
            off_screen_plotter.enable_depth_peeling()
        
        # copy the mesh and camera settings to the off-screen plotter
        for actor in self.plotter.renderer.actors.values():
            off_screen_plotter.add_actor(actor)
        
        off_screen_plotter.camera_position = self.plotter.camera_position
        off_screen_plotter.enable_anti_aliasing('msaa', multi_samples=16)

        screenshot_extensions = ['png', 'jpeg', 'jpg', 'bmp', 'tif', 'tiff']
        if file_name.split('.')[-1] in screenshot_extensions:
            off_screen_plotter.screenshot(file_name)
        else:
            off_screen_plotter.save_graphic(file_name, title='GeViewer Figure')

        del off_screen_plotter
        if self.success:
            self.print_to_console('Success: figure saved to ' + file_name)
        else:
            self.print_to_console('Error: figure not saved.')
        self.success = True

    ##############################################################
    # Methods for the tools tab
    ##############################################################

    def check_geometry(self, tolerance, samples):
        """Checks the geometry.

        This method checks the geometry by finding the overlaps between
        the components and updating the overlaps list.

        :param tolerance: The tolerance for the overlaps.
        :type tolerance: float
        :param samples: The number of samples to use for the overlaps.
        :type samples: int
        """
        if self.worker_running:
            self.print_to_console('Error: wait for the current process to finish.')
            return
        tolerance = float(tolerance) if tolerance else 0.001
        samples = int(samples) if samples else 10000
        if not len(self.viewer.components):
            self.print_to_console('Error: no components loaded.')
            return
        self.print_to_console('Checking {} for overlaps...'.format(self.viewer.components[0]['name']))
        self.worker = Worker(self.viewer.find_overlaps, self.progress_bar, tolerance=tolerance, n_samples=samples)
        self.worker.on_finished(self.show_overlaps)
        self.worker.error_signal.connect(self.global_exception_hook)
        self.worker_running = True
        self.worker.start()


    def show_overlaps(self):
        """Shows the overlaps.

        This method shows the overlaps by setting the checkboxes to checked
        for the overlapping meshes and toggling the transparency if it is not
        already enabled.
        """
        overlapping_meshes = self.worker.get_result()
        self.worker.deleteLater()
        self.worker_running = False
        if self.success:
            if len(overlapping_meshes) == 0:
                self.print_to_console('Success: no overlaps found.')
            else:
                for checkbox in self.checkbox_mapping.values():
                    checkbox.setCheckState(Qt.CheckState.Unchecked)
                for mesh_id in overlapping_meshes:
                    self.checkbox_mapping[mesh_id].setCheckState(Qt.CheckState.Checked)
                if not self.viewer.transparent:
                    self.viewer.toggle_transparent()
                self.plotter.view_isometric()
                self.print_to_console('Found {} potential overlaps.'.format(len(overlapping_meshes)))
        else:
            self.print_to_console('Error: overlap check failed.')
        self.success = True
        self.progress_bar.signal_finished()
        self.progress_bar.interrupt = False


    def clear_overlaps(self):
        """Clears the overlaps.

        This method clears the overlaps by removing all actors from the plotter
        and clearing the overlaps list.
        """
        for actor in self.viewer.overlaps:
            self.plotter.remove_actor(actor)
        self.viewer.overlaps = []
        self.print_to_console('Overlaps cleared.')


    def measure_distance(self):
        """Measures the distance.

        This method measures the distance by adding a measurement widget to the plotter
        and clearing the measurement box.
        """
        self.print_to_console('Measuring distance. Click on two points to measure the distance between them.')
        self.plotter.add_measurement_widget(self.display_measurement)


    def display_measurement(self, point1, point2, distance):
        """Displays the measurement.

        This method displays the measurement by setting the measurement box
        to the distance between the two points.

        :param point1: The first point.
        :type point1: list
        :param point2: The second point.
        :type point2: list
        :param distance: The distance between the two points.
        :type distance: float
        """
        boxes = [self.measurement_box, self.measurement_box_2, self.measurement_box_3]
        for box in boxes:
            if box.text() == '':
                box.setText('{:.3f}'.format(distance))
                return
        boxes[0].setText(boxes[1].text())
        boxes[1].setText(boxes[2].text())
        boxes[2].setText('{:.3f}'.format(distance))


    def clear_measurement(self, print_to_console=True):
        """Clears the measurement.

        This method clears the measurement by clearing the measurement boxes.
        """
        self.plotter.clear_measure_widgets()
        self.measurement_box.setText('')
        self.measurement_box_2.setText('')
        self.measurement_box_3.setText('')
        if print_to_console:
            self.print_to_console('Measurements cleared.')

    ##############################################################
    # Methods for the console
    ##############################################################

    def print_to_console(self, text):
        """Prints to the console.

        This method prints text to the console with the GeViewer prompt.

        :param text: The text to print to the console.
        :type text: str
        """
        print('[geviewer-prompt]: ' + text)

    ##############################################################
    # Methods for the menu bar
    ##############################################################

    def open_file_dialog(self):
        """Opens the file dialog.

        This method opens the file dialog by getting the open file name
        and emitting the file name changed event.
        """
        try:
            options = QFileDialog.Options()
            dialog = QFileDialog(self, 'Open File', '', 
                                 'All Supported Files (*.wrl *.heprep *.gev);;' +
                                 'VRML Files (*.wrl);;HepRep Files (*.heprep);;'
                                 'GEV Files (*.gev)', options=options)
            dialog.setFileMode(QFileDialog.ExistingFile)
            dialog.fileSelected.connect(self.load_file)
            dialog.show()
        except Exception as e:
            self.global_exception_hook(type(e), e, e.__traceback__)


    def load_file(self, file_path):
        """Loads the file.
        
        :param file_path: The path to the file to load.
        :type file_path: str
        """
        if file_path:
            if self.worker_running:
                self.print_to_console('Error: wait for the current process to finish.')
                return
            start_time = time.time()
            self.progress_bar.setValue(0)
            self.file_name_changed.emit(file_path)
            self.print_to_console('Loading file: {}\n'.format(file_path))
            self.worker = Worker(self.load_and_plot, self.progress_bar, filename=file_path)
            self.worker.finished.connect(lambda: self.on_file_loaded(start_time))
            self.worker.error_signal.connect(self.global_exception_hook)
            self.worker.finished.connect(self.worker.deleteLater)
            self.worker_running = True
            self.worker.start()


    def load_and_plot(self, progress_obj, filename):
        """The function to call when a file is loaded.

        :param progress_obj: The progress object to use.
        :type progress_obj: ProgressBar
        :param filename: The path to the file to load.
        :type filename: str
        """
        self.viewer.load_file(filename=filename, progress_obj=progress_obj, off_screen=True)
        self.viewer.create_plotter(progress_obj=progress_obj)


    def on_file_loaded(self, start_time=None):
        """Method to call when a file is loaded.

        :param start_time: The start time of the file loading.
        :type start_time: float, optional
        """
        self.worker_running = False
        self.generate_checkboxes(self.viewer.components, 0)
        time_str = ''
        time_elapsed = 0
        if start_time:
            time_elapsed = time.time() - start_time
            time_str = ' in {:.2f} seconds'.format(time_elapsed)
        if self.success:
            self.print_to_console('Success: file loaded' + time_str + '.')
            if time_elapsed > 10:
                self.print_to_console('Hint: save as .gev to speed up loading next time.')
        else:
            self.print_to_console('Error: file could not be loaded.')
        self.success = True
        self.progress_bar.signal_finished()
        self.progress_bar.interrupt = False
        self.file_loaded.emit()


    def save_file_dialog(self):
        """Saves the file.

        This method saves the file by getting the save file name and emitting
        the save file event.
        """
        try:
            options = QFileDialog.Options()
            file_types = 'GeViewer File (*.gev);; All Files (*)'
            file_name, _ = QFileDialog.getSaveFileName(self, 'Save File', 'viewer.gev', file_types, options=options)
            
            if file_name:
                self.viewer.save_session(file_name)
                if self.success:
                    self.print_to_console('Success: file saved to {}.'.format(file_name))
                else:
                    self.print_to_console('Error: file could not be saved.')
                self.success = True
                
        except Exception as e:
            self.global_exception_hook(type(e), e, e.__traceback__)


    def clear_viewer(self):
        """Clears the meshes.

        This method clears the meshes by removing all actors from the plotter
        and clearing the checkbox mapping.
        """
        self.viewer.clear_meshes()
        self.plotter.reset_camera()
        for checkbox in self.checkbox_mapping.values():
            checkbox.setParent(None)
            checkbox.deleteLater()
        self.checkbox_mapping.clear()
        while self.checkboxes_layout.count() > 0:
            item = self.checkboxes_layout.takeAt(0)
            item.widget().deleteLater()
        self.checkboxes_layout.addWidget(self.load_instructions)
        self.load_instructions.setParent(self.checkboxes_layout.parentWidget())
        self.load_instructions.show()
        self.events_list.clear()
        self.current_file.clear()
        self.file_name_changed.emit(None)
        self.event_selection_box.setRange(1, 1)
        self.event_selection_box.setValue(1)
        self.clear_measurement(print_to_console=False)
        gc.collect()
        self.print_to_console('Viewer cleared.')


    def abort_process(self):
        """Aborts a running process.
        """
        if self.worker_running:
            self.print_to_console('Aborting current process...')
            self.success = False
            self.progress_bar.interrupt_worker()
        else:
            self.print_to_console('No process to abort.')


    def toggle_components_panel(self):
        """Toggles the components panel.

        This method toggles the components panel by showing or hiding it
        depending on its current state.
        """
        if self.components_panel.isVisible():
            self.components_panel.hide()
            self.print_to_console('Hiding components panel.')
        else:
            self.components_panel.show()
            self.print_to_console('Showing components panel.')


    def toggle_control_panel(self):
        """Toggles the control panel.

        This method toggles the control panel by showing or hiding it
        depending on its current state.
        """
        if self.control_panel.isVisible():
            self.control_panel.hide()
            self.print_to_console('Hiding control panel.')
        else:
            self.control_panel.show()
            self.print_to_console('Showing control panel.')


    def toggle_background(self):
        """Toggles the background.

        This method toggles the background by setting the background attribute
        of the viewer to the opposite of its current value and updating the
        background button.
        """
        self.print_to_console('Toggling background ' + ['on.','off.'][self.viewer.bkg_on])
        self.viewer.toggle_background()
        self.viewer.plotter.update()


    def toggle_gradient(self):
        """Toggles the gradient.

        This method toggles the gradient by setting the gradient attribute
        of the viewer to the opposite of its current value and updating the
        background color of the viewer.
        """
        self.print_to_console('Toggling gradient ' + ['on.','off.'][self.viewer.gradient])
        self.viewer.gradient = not self.viewer.gradient
        self.viewer.set_background_color()


    def toggle_depth_peeling(self, on=False):
        """Toggles depth peeling.

        This method turns on or off depth peeling for the plotter.
        """
        self.print_to_console('Turning {} depth peeling.'.format(['off','on'][on]))
        if on:
            self.plotter.enable_depth_peeling()
        else:
            self.plotter.disable_depth_peeling()


    def open_color_picker(self, button):
        """Opens the color picker.

        This method opens the color picker by getting the color from the color
        dialog and setting the background color of the viewer.

        :param button: The button to open the color picker for.
        :type button: int
        """
        if button == 2:
            self.viewer.bkg_colors = ['lightskyblue', 'midnightblue']
            self.print_to_console('Resetting background color.')
        else:
            color = QColorDialog.getColor()
            if color.isValid():
                self.viewer.bkg_colors[button] = color.getRgbF()[:3]
                hex_color = color.name()
                self.print_to_console('Setting background color {} to {}.'.format(button + 1, hex_color))
        self.viewer.set_background_color()


    def show_license(self):
        """Shows the license.

        This method shows the license by reading the license file and
        printing it to the console.
        """
        with open('LICENSE') as f:
            license_text = f.read().replace('\n\n', '<>').replace('\n', ' ').replace('<>', '\n\n')
        self.print_to_console('\n' + license_text)


    def check_for_updates(self):
        """Checks for updates.

        This method checks for updates by calling the check_for_updates function
        from the utils module and printing the result to the console.
        """
        updates = utils.check_for_updates()
        if updates:
            self.print_to_console(updates)
        else:
            self.print_to_console('You are using the latest version of GeViewer.')


    def show_documentation(self):
        """Shows the documentation.

        This method shows the documentation by opening the documentation in
        the default web browser.
        """
        try:
            webbrowser.open('https://geviewer.readthedocs.io/en/latest/')
        except Exception as e:
            doc_url = 'https://geviewer.readthedocs.io/en/latest/'
            self.print_to_console('Find the GeViewer documentation at ' + \
                '<a href="' + doc_url + '">' + doc_url + '</a>')


class ConsoleRedirect(io.StringIO):
    """Redirects stdout and stderr to a QTextEdit widget.
    """
    def __init__(self, console):
        """Initializes the console redirect.

        :param text_edit: The text edit widget to redirect to.
        :type text_edit: QTextEdit
        """
        super().__init__()
        self.console = console


    def write(self, text):
        """Writes to the console.

        This method writes to the console by inserting the text into the text
        edit widget and moving the cursor to the end.

        :param text: The text to write to the console.
        :type text: str
        """
        self.console.moveCursor(QTextCursor.End)
        text = self.stylize_text(text)
        self.console.insertHtml(text)
        self.console.moveCursor(QTextCursor.End)
        super().write(text)

        
    def stylize_text(self, text):
        """Stylizes the text.

        This method stylizes the text by adding a prompt to the beginning of
        the text, replacing newlines with HTML line breaks, and replacing
        warnings, errors, and successes with HTML formatted text.

        :param text: The text to stylize.
        :type text: str
        :return: The stylized text.
        :rtype: str
        """
        prompt = QDateTime.currentDateTime().toString('[yyyy-MM-dd HH:mm:ss]: ')
        text = text.replace('[geviewer-prompt]: ', '<b style="color: blue;">{}</b>'.format(prompt))
        text = text.replace('\n', '<br>')
        text = re.sub(r'\b(Warning)\b', r'<b style="color: orange;">\1</b>', text)
        text = re.sub(r'\b(Error)\b', r'<b style="color: red;">\1</b>', text)
        text = re.sub(r'\b(Success)\b', r'<b style="color: green;">\1</b>', text)
        text = re.sub(r'\b(Hint)\b', r'<b style="color: purple;">\1</b>', text)
        return text


    def flush(self):
        pass

    
class ProgressBar(QProgressBar):
    """A custom progress bar class for the GeViewer application.

    This class handles all interactions between the worker threads and
    the GUI, including updating the progress bar, sending text updates,
    and interrupting the worker threads if necessary.
    """
    #: Signal emitted to update the progress value
    #:
    #: :param int: The new progress value
    progress = pyqtSignal(int)

    #: Signal emitted to set the maximum value of the progress bar
    #:
    #: :param int: The new maximum value
    maximum = pyqtSignal(int)

    #: Signal emitted when the task is finished
    finished = pyqtSignal()

    #: Signal emitted to send status updates
    #:
    #: :param str: The update text
    update = pyqtSignal(str)

    #: Signal emitted to start or stop the timer
    #:
    #: :param bool: True to start the timer, False to stop
    run_timer = pyqtSignal(bool)

    def __init__(self):
        """Initializes the progress bar.
        """
        super().__init__()
        self.setValue(0)
        self.setRange(0, 100)
        self._internal_value = 0
        self.current_value = 0
        self.maximum_value = 100
        self.update_buffer = []
        self.new_progress = False
        self.interrupt = False
        self.progress.connect(self.setValue, Qt.QueuedConnection)
        self.maximum.connect(self.setMaximum, Qt.QueuedConnection)
        self.finished.connect(lambda: self.setValue(self.maximum_value), Qt.QueuedConnection)
        self.update.connect(print)
        self.run_timer.connect(self.start_or_stop_timer)
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.send_updates)


    def increment_progress(self):
        """Increments the progress bar by 1%.
        """
        self.current_value += 1
        if 100*(self.current_value - self._internal_value)/self.maximum_value >= 1:
            self._internal_value = self.current_value
            self.new_progress = True

            
    def reset_progress(self):
        """Resets the progress bar to 0%.
        """
        self.current_value = 0
        self._internal_value = 0
        self.progress.emit(0)
        self.run_timer.emit(True)


    def set_maximum_value(self, value):
        """Sets the maximum value of the progress bar.

        :param value: The maximum value of the progress bar.
        :type value: int
        """
        self.maximum_value = value
        self.maximum.emit(value)


    def signal_finished(self):
        """Signals that the progress bar has finished.
        """
        self.send_updates()
        self.run_timer.emit(False)
        self.finished.emit()


    def start_or_stop_timer(self, start):
        """Starts or stops the timer in the main thread.

        :param start: Whether to start the timer.
        :type start: bool
        """
        if start:
            self.timer.start(500)
        else:
            self.timer.stop()


    def print_update(self, text):
        """Adds an update to the update buffer.
        """
        if self.timer.isActive():
            self.update_buffer.append(text)
        else:
            self.update.emit(text)


    def sync_status(self, update=None, increment=False):
        """Synchronizes the status of the worker with that of the user interface.
        This should be called in the following format:

        .. code-block:: python

            pbar = ProgressBar()
            if pbar.sync_status(update, increment): return
        
        where `update` is a string and `increment` is a boolean.
        This will pass the status updates to the user interface and return True
        if the worker should be interrupted.

        :param update: The update to send to the user interface.
        :type update: str, optional
        :param increment: Whether to increment the progress bar.
        :type increment: bool, optional
        :return: Whether the worker should be interrupted.
        :rtype: bool
        """
        if self.interrupt: return True
        if update:
            self.print_update(update)
        if increment:
            self.increment_progress()


    def interrupt_worker(self):
        """Flags an interrupt to the worker.
        """
        if self.timer.isActive():
            self.interrupt = True


    def send_updates(self):
        """Flushes the update buffer.
        """
        if self.new_progress:
            self.progress.emit(self._internal_value)
            self.new_progress = False
        if self.update_buffer:
            self.update.emit('\n'.join(self.update_buffer))
            self.update_buffer.clear()
        

class Worker(QThread):
    """A custom worker class for the GeViewer application.
    """
    #: Signal emitted when the worker has finished its task
    finished = pyqtSignal()

    #: Signal emitted when an error occurs during the worker's task
    #:
    #: :param type: The type of the exception
    #: :param Exception: The exception instance
    #: :param object: The traceback object
    error_signal = pyqtSignal(type, Exception, object)

    def __init__(self, task, progress_bar, **kwargs):
        """Initializes the worker.

        :param task: The task to be executed by the worker
        :type task: callable
        :param progress_bar: The progress bar to update during the task
        :type progress_bar: ProgressBar
        :param kwargs: Additional keyword arguments for the task
        """
        super().__init__()
        self.task = task
        self.kwargs = kwargs
        self.progress_bar = progress_bar
        self.result = None


    def run(self):
        """Runs the worker.
        """
        try:
            self.result = self.task(progress_obj=self.progress_bar, **self.kwargs)
        except Exception as e:
            self.error_signal.emit(type(e), e, e.__traceback__)
        finally:
            self.finished.emit()


    def on_finished(self, func):
        """Connects a function to the finished signal.

        :param func: The function to connect to the finished signal.
        :type func: function
        """
        self.finished.connect(func)


    def get_result(self):
        """Returns the result of the worker.
        """
        return self.result


    def cleanup(self):
        """Cleans up the resources used by the worker.
        """
        self.task = None
        self.kwargs.clear()
        self.progress_bar = None
        self.result = None


    def deleteLater(self):
        """Overrides the deleteLater method to clean up the resources used by the worker.
        """
        self.cleanup()
        super().deleteLater()


def launch_app(files_to_load=None):
    """Launches the app.
    """
    app = Application([])
    app.setStyle('Fusion')
    
    window = Window()
    sys.excepthook = window.global_exception_hook
    window.show()

    if files_to_load:
        window.load_initial_files(files_to_load)

    return app.exec()



import unittest
from unittest import mock
import os
import tempfile
from geviewer import converter, viewer
from pathlib import Path

class TestGevConverter(unittest.TestCase):

    @mock.patch('argparse.ArgumentParser.parse_args')
    def test_gev_converter_wrl(self, mock_parse_args):
        """Tests the gev-converter command-line utility with a VRML file.
        """

        with tempfile.TemporaryDirectory() as temp_dir:
            input_file = 'tests/sample.wrl'
            output_file = os.path.join(temp_dir, 'sample.gev')
            
            mock_args = mock.Mock()
            mock_args.file = Path(input_file)
            mock_args.destination = Path(output_file)
            mock_parse_args.return_value = mock_args

            converter.main()
            
            gev = viewer.GeViewer()
            gev.load_file(output_file, off_screen=True)
            self.assertEqual(len(gev.components[0]['children']), 3)


    @mock.patch('argparse.ArgumentParser.parse_args')
    def test_gev_converter_heprep(self, mock_parse_args):
        """Tests the gev-converter command-line utility with a HepRep file.
        """
        
        with tempfile.TemporaryDirectory() as temp_dir:
            input_file = 'tests/sample.heprep'
            output_file = os.path.join(temp_dir, 'sample.gev')

            mock_args = mock.Mock()
            mock_args.file = Path(input_file)
            mock_args.destination = Path(output_file)
            mock_parse_args.return_value = mock_args

            converter.main()
            
            gev = viewer.GeViewer()
            gev.load_file(output_file, off_screen=True)
            self.assertEqual(len(gev.components[0]['children']), 2)


if __name__ == '__main__':
    unittest.main()


import numpy as np
import pyvista as pv
import os
import shutil
import zipfile
import tempfile
import json
import gc
from pyvistaqt import QtInteractor
from geviewer import parsers


class GeViewer:
    """The main interface for the GeViewer application, responsible for loading,
    processing, and visualizing data files. This class manages the creation
    and display of 3D visualizations based on the provided data files and
    offers various functionalities such as toggling display options and saving
    sessions.
    """

    def __init__(self, plotter_widget=None):
        """Initializes the GeViewer object.

        :param plotter_widget: The widget to use for the plotter.
        :type plotter_widget: QWidget, optional
        """
        self.off_screen = False
        if plotter_widget:
            self.plotter = QtInteractor(plotter_widget)
        else:
            self.plotter = pv.Plotter()
        self.bkg_colors = ['lightskyblue', 'midnightblue']
        self.plotter.set_background(*self.bkg_colors)
        self.bkg_on = True
        self.wireframe = False
        self.transparent = False
        self.gradient = True
        self.parallel = False
        self.components = []
        self.overlaps = []
        self.event_ids = []
        self.actors = {}


    def load_file(self, filename, off_screen=False, progress_obj=None):
        """Loads the file into the components list.

        :param filename: The name of the file to load.
        :type filename: str
        :param off_screen: If True, the plotter is created without displaying it. Defaults to False.
        :type off_screen: bool, optional
        :param progress_obj: The progress bar object to use.
        :type progress_obj: ProgressBar, optional
        """
        self.off_screen = off_screen
        if filename.endswith('.gev'):
            new_components = self.load_session(filename)
        elif filename.endswith('.wrl'):
            parser = parsers.VRMLParser(filename)
            parser.parse_file(progress_obj)
            new_components = [parser.components]
        elif filename.endswith('heprep'):
            parser = parsers.HepRepParser(filename)
            parser.parse_file(progress_obj)
            new_components = parser.components
        self.num_to_plot = self.count_components(new_components)
        self.components.extend(new_components)


    def count_components(self, components, exclude_events=False, exclude_invisible=False):
        """Counts the number of components in the list of components.

        :param components: A list of components.
        :type components: list
        :return: The number of components in the list.
        :rtype: int
        """
        if self.off_screen:
            exclude_invisible = False
        count = 0
        for comp in components:
            if not (exclude_events and ((comp['mesh'] is None) or (comp['shape'] == 'Point') or \
                (comp['shape'] == 'Line')) or (exclude_invisible and not self.actors[comp['id']].GetVisibility())):
                count += 1
            if len(comp['children']) > 0:
                count += self.count_components(comp['children'], exclude_events=exclude_events, \
                                               exclude_invisible=exclude_invisible)
        return count
    
    
    def create_plotter(self, progress_obj=None):
        """Creates the plotter and plots the meshes.

        :param progress_obj: The progress object to use for the plotter.
        :type progress_obj: ProgressBar, optional
        """
        update = 'Plotting meshes...\n'
        if progress_obj:
            progress_obj.reset_progress()
            progress_obj.set_maximum_value(self.num_to_plot)
            if progress_obj.sync_status(update=update): return
        else:
            print(update)
        self.plot_meshes(self.components, progress_obj=progress_obj)
        if progress_obj:
            progress_obj.signal_finished()


    def plot_meshes(self, components, level=0, progress_obj=None):
        """Plots the meshes and saved the actors in a dictionary.

        :param components: The components to plot.
        :type components: list
        :param level: Keeps track of the recursion level.
        :type level: int, optional
        :param progress_obj: The progress object to use for the plotter.
        :type progress_obj: ProgressBar, optional
        """
        style = 'wireframe' if self.wireframe else 'surface'
        opacity = 0.3 if self.transparent else 1.
        for comp in components:
            if comp['mesh'] is not None and not comp['has_actor']:
                update = '...'*level + 'Plotting ' + comp['name'] + '...\n'
                if progress_obj:
                    if progress_obj.sync_status(update=update): return
                else:
                    print(update)
                if comp['is_event']:
                    self.event_ids.append(comp['id'])
                    this_opacity = 1
                else:
                    this_opacity = opacity
                actor = self.plotter.add_mesh(comp['mesh'], scalars='color', rgb=True, \
                                              render_points_as_spheres=comp['is_dot'], \
                                              point_size=5*comp['is_dot'], style=style, \
                                              opacity=this_opacity, name=comp['id'])
                if comp['name'] == '/nEXO/TPCInternals/LXe/ActiveRegion':
                    self.bad_actor = comp['mesh']
                self.actors[comp['id']] = actor
                comp['has_actor'] = True
                if progress_obj:
                    if progress_obj.sync_status(increment=True): return
            if len(comp['children']) > 0:
                self.plot_meshes(comp['children'], level + 1, progress_obj)
        if level == 0:
            self.plotter.view_isometric()
            update = 'Done plotting.\n'
            if progress_obj:
                if progress_obj.sync_status(update=update): return
            else:
                print(update)
            self.num_to_plot = 0
        

    def set_background_color(self):
        """Sets the background color.
        """
        if not self.bkg_on:
            return
        if self.gradient:
            self.plotter.set_background(self.bkg_colors[0],top=self.bkg_colors[1])
        else:
            self.plotter.set_background(self.bkg_colors[0])


    def toggle_parallel_projection(self):
        """Toggles the parallel projection on and off.
        """
        if not self.parallel:
            self.plotter.enable_parallel_projection()
        else:
            self.plotter.disable_parallel_projection()
        self.parallel = not self.parallel


    def toggle_background(self):
        """Toggles the gradient background on and off.
        """
        self.bkg_on = not self.bkg_on
        if self.bkg_on:
            top = self.bkg_colors[1] if self.gradient else None
            self.plotter.set_background(self.bkg_colors[0],top=top)
        else:
            self.plotter.set_background('white')
        if not self.off_screen:
            self.plotter.update()


    def toggle_wireframe(self):
        """Toggles between solid and wireframe display modes. Disables depth
        peeling if wireframe mode is enabled to improve responsiveness.
        """
        self.wireframe = not self.wireframe
        if self.wireframe:
            for actor in self.actors.values():
                actor.GetProperty().SetRepresentationToWireframe()
        else:
            for actor in self.actors.values():
                actor.GetProperty().SetRepresentationToSurface()
        if not self.off_screen:
            self.plotter.update()


    def toggle_transparent(self):
        """Toggles transparency on and off.
        """
        self.transparent = not self.transparent
        if self.transparent:
            for id, actor in self.actors.items():
                if id in self.event_ids:
                    continue
                actor.GetProperty().SetOpacity(0.3)
        else:
            for actor in self.actors.values():
                actor.GetProperty().SetOpacity(1)
        if not self.off_screen:
            self.plotter.update()


    def save_session(self, filename):
        """Saves the session to a .gev file.

        :param filename: The name of the file to save the session to.
        :type filename: str
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpfolder = tmpdir + '/gevfile/'
            os.makedirs(tmpfolder, exist_ok=False)
            def save_serializable_entries(components, level=0, saveable_dicts=[]):
                for comp in components:
                    temp_dict = {}
                    for key, value in comp.items():
                        if key not in ['mesh_points', 'mesh_inds', 'scalars', 'mesh', 'actor']:
                            temp_dict[key] = value
                    if comp['mesh_points'] is not None:
                        np.save(tmpfolder + 'mesh_points_{}.npy'.format(comp['id']), comp['mesh_points'], allow_pickle=False)
                        temp_dict['mesh_points'] = 'mesh_points_{}.npy'.format(comp['id'])
                    else:
                        temp_dict['mesh_points'] = None
                    if comp['mesh_inds'] is not None:
                        np.save(tmpfolder + 'mesh_inds_{}.npy'.format(comp['id']), comp['mesh_inds'], allow_pickle=False)
                        temp_dict['mesh_inds'] = 'mesh_inds_{}.npy'.format(comp['id'])
                    else:
                        temp_dict['mesh_inds'] = None
                    if comp['scalars'] is not None:
                        np.save(tmpfolder + 'scalars_{}.npy'.format(comp['id']), comp['scalars'], allow_pickle=False)
                        temp_dict['scalars'] = 'scalars_{}.npy'.format(comp['id'])
                    else:
                        temp_dict['scalars'] = None
                    if comp['mesh'] is not None:
                        comp['mesh'].save(tmpfolder + 'mesh_{}.vtk'.format(comp['id']))
                        temp_dict['mesh'] = 'mesh_{}.vtk'.format(comp['id'])
                    else:
                        temp_dict['mesh'] = None
                    temp_dict['has_actor'] = False
                    if len(comp['children']) > 0:
                        temp_dict['children'] = []
                        save_serializable_entries(comp['children'], level + 1, temp_dict['children'])
                    saveable_dicts.append(temp_dict)
                if level == 0:
                    return saveable_dicts
                
            saveable_dicts = save_serializable_entries(self.components)

            for i, saveable_dict in enumerate(saveable_dicts):
                with open(tmpfolder + 'components_dict_{}.json'.format(i), 'w') as f:
                    json.dump(saveable_dict, f)

            with zipfile.ZipFile(tmpdir + '/gevfile.gev', 'w') as archive:
                for file_name in os.listdir(tmpfolder):
                    file_path = os.path.join(tmpfolder, file_name)
                    archive.write(file_path, arcname=file_name)

            # if using the default filename and it exists, increment
            # the number until a unique filename is found
            if filename=='viewer.gev' and os.path.exists(filename):
                filename = 'viewer2.gev'
                i = 2
                while(os.path.exists('viewer{}.gev'.format(i))):
                    i += 1
                filename = 'viewer{}.gev'.format(i)
            shutil.copy(tmpdir + '/gevfile.gev', filename)

                
    def load_session(self, filename):
        """Loads the session from a .gev file.

        :param filename: The name of the file to load the session from.
        :type filename: str
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpfolder = tmpdir + '/gevfile/'
            os.makedirs(tmpfolder, exist_ok=False)
            with zipfile.ZipFile(filename, 'r') as archive:
                archive.extractall(tmpfolder)
            files = [f for f in os.listdir(tmpfolder) if f.endswith('.json')]
            components = []
            for file in files:
                with open(tmpfolder + file, 'r') as f:
                    comp = json.load(f)
                    components.append(comp)

            def load_components(components, level=0):
                for comp in components:
                    if comp['mesh_points'] is not None:
                        comp['mesh_points'] = np.load(tmpfolder + comp['mesh_points'], allow_pickle=False)
                    if comp['mesh_inds'] is not None:
                        comp['mesh_inds'] = np.load(tmpfolder + comp['mesh_inds'], allow_pickle=False)
                    if comp['scalars'] is not None:
                        comp['scalars'] = np.load(tmpfolder + comp['scalars'], allow_pickle=False)
                    if comp['mesh'] is not None:
                        comp['mesh'] = pv.read(tmpfolder + comp['mesh'])
                    if len(comp['children']) > 0:
                        load_components(comp['children'], level + 1)
            
            load_components(components)
            return components
        
        
    def is_mesh_inside(self, mesh1, mesh2):
        """Checks if one mesh is inside another.

        :param mesh1: The first mesh.
        :type mesh1: pyvista.PolyData
        :param mesh2: The second mesh.
        :type mesh2: pyvista.PolyData
        :return: True if mesh1 is inside mesh2, False otherwise.
        :rtype: bool
        """
        bounds1 = mesh1.bounds
        bounds2 = mesh2.bounds
        if bounds1[0] >= bounds2[0] and bounds1[1] <= bounds2[1] and \
           bounds1[2] >= bounds2[2] and bounds1[3] <= bounds2[3] and \
           bounds1[4] >= bounds2[4] and bounds1[5] <= bounds2[5]:
            return True
        return False
    

    def do_bounds_overlap(self, mesh1, mesh2):
        """Checks if the bounds of two meshes overlap in all three dimensions.

        :param mesh1: The first mesh.
        :type mesh1: pyvista.PolyData
        :param mesh2: The second mesh.
        :type mesh2: pyvista.PolyData
        :return: True if the bounds overlap in all dimensions, False otherwise.
        :rtype: bool
        """
        bounds1 = mesh1.bounds
        bounds2 = mesh2.bounds
        x_overlap = bounds1[0] <= bounds2[1] and bounds2[0] <= bounds1[1]
        y_overlap = bounds1[2] <= bounds2[3] and bounds2[2] <= bounds1[3]
        z_overlap = bounds1[4] <= bounds2[5] and bounds2[4] <= bounds1[5]
        
        return x_overlap and y_overlap and z_overlap
    
    
    def get_overlap(self, mesh1, mesh2, tolerance=0.001, n_samples=100000, progress_obj=None):
        """Gets the overlap between two meshes.

        :param mesh1: The first mesh.
        :type mesh1: pyvista.PolyData
        :param mesh2: The second mesh.
        :type mesh2: pyvista.PolyData
        :param tolerance: The tolerance for the overlap.
        :type tolerance: float, optional
        :param n_samples: The number of samples to use.
        :type n_samples: int, optional
        :return: The points of overlap and the fraction of points that survived.
        :rtype: tuple
        """

        # get the disparate regions in mesh 1
        connected_components_1 = mesh1.connectivity()
        num_regions_1 = connected_components_1['RegionId'].max()
        if num_regions_1 == 0:
            separated_meshes_1 = [mesh1]
        else:
            separated_meshes_1 = []
            for i in range(num_regions_1):
                region_mesh = connected_components_1.threshold([i, i], scalars="RegionId")
                separated_meshes_1.append(region_mesh)

        # get the disparate regions in mesh 2
        connected_components_2 = mesh2.connectivity()
        num_regions_2 = connected_components_2['RegionId'].max()
        if num_regions_2 == 0:
            separated_meshes_2 = [mesh2]
        else:
            separated_meshes_2 = []
            for i in range(num_regions_2):
                region_mesh = connected_components_2.threshold([i, i], scalars="RegionId")
                separated_meshes_2.append(region_mesh)

        total_checks = len(separated_meshes_1)*len(separated_meshes_2)

        # check for overlaps between all regions in mesh 1 and all regions in mesh 2
        points = []
        n_surviving = 0
        current_check = 0
        for mesh1 in separated_meshes_1:
            for mesh2 in separated_meshes_2:

                if total_checks > 200 and current_check % 100 == 0:
                    update = 'Starting check {}/{}...{}'.format(current_check + 1, total_checks, \
                                                                ['','\n'][total_checks - current_check < 100])
                    if progress_obj:
                        if progress_obj.sync_status(update=update): return
                    else:
                        print(update)

                mesh1 = mesh1.extract_surface()
                mesh2 = mesh2.extract_surface()

                if not self.do_bounds_overlap(mesh1, mesh2):
                    continue

                mc_points = np.random.uniform(low=mesh1.bounds[::2], \
                                              high=mesh1.bounds[1::2], \
                                              size=(n_samples, 3))
                
                mc_points = pv.PolyData(mc_points)
                select = mc_points.select_enclosed_points(mesh1, tolerance=1e-6)
                mc_points = select.points[select['SelectedPoints'].astype(bool)]
                n_surviving += mc_points.shape[0]
                mc_points = pv.PolyData(mc_points)

                select = mc_points.select_enclosed_points(mesh2, tolerance=1e-6)
                mc_points = select.points[select['SelectedPoints'].astype(bool)]
                mc_points = pv.PolyData(mc_points)
                select = mc_points.compute_implicit_distance(mesh2)
                bounds = mesh2.bounds
                dimensions = np.array([bounds[1] - bounds[0], bounds[3] - bounds[2], bounds[5] - bounds[4]])
                mc_points = select.points[np.abs(select['implicit_distance']) > tolerance*np.linalg.norm(dimensions)]
                points.append(mc_points)

                current_check += 1

        if len(points) > 0:
            points = pv.PolyData(np.concatenate(points))
        else:
            points = pv.PolyData()

        overlap_fraction = None if n_surviving == 0 else points.n_points/n_surviving

        return points, overlap_fraction
        
        
    def find_overlaps(self, tolerance=0.001, n_samples=100000, progress_obj=None):
        """Finds the overlaps between the meshes.

        :param tolerance: The tolerance for the overlap.
        :type tolerance: float, optional
        :param n_samples: The number of samples to use.
        :type n_samples: int, optional
        :return: The ids of the meshes that overlap.
        :rtype: list
        """
        for actor in self.overlaps:
            self.plotter.remove_actor(actor)
        self.overlaps.clear()
        overlapping_meshes = []
        checked = []

        def find_overlaps_recursive(components, level=0, progress_obj=None):
            """Finds the overlaps between the meshes.

            :param components: The components to check for overlaps.
            :type components: list
            :param level: Keeps track of the recursion level.
            :type level: int, optional
            :param progress_obj: The progress bar object to use.
            :type progress_obj: ProgressBar, optional
            """
            for comp in components:
                if comp['mesh'] is not None and not (comp['shape'] == 'Point' or comp['shape'] == 'Line'):
                    check_for_overlaps(comp, self.components[0]['children'], progress_obj)
                if len(comp['children']) > 0:
                    find_overlaps_recursive(comp['children'], level + 1, progress_obj)
                checked.append(comp['id'])

        def check_for_overlaps(comp1, components, progress_obj=None):
            """Checks for overlaps between one component and all other components.

            :param comp1: The first component.
            :type comp1: dict
            :param components: The components to check for overlaps.
            :type components: list
            :param progress_obj: The progress bar object to use.
            :type progress_obj: ProgressBar, optional
            """
            for comp2 in components:

                if comp2['mesh'] is not None and not (comp2['shape'] == 'Point' or comp2['shape'] == 'Line') \
                    and (comp1['id'] != comp2['id']) and (comp1['id'] not in checked) and (comp2['id'] not in checked):

                    mesh1 = comp1['mesh']
                    mesh2 = comp2['mesh']

                    skip = False
                    if not self.off_screen and not (self.actors[comp1['id']].GetVisibility() and \
                                                    self.actors[comp2['id']].GetVisibility()):
                        skip = True
                    if not skip:
                        update = 'Checking {} and {}...\n'.format(comp1['name'], comp2['name'])
                        if progress_obj:
                            if progress_obj.sync_status(update=update, increment=True): return
                        else:
                            print(update)
                    if not skip and not mesh1.is_all_triangles:
                        mesh1 = mesh1.triangulate()
                    if not skip and not mesh2.is_all_triangles:
                        mesh2 = mesh2.triangulate()
                    if not skip and (self.is_mesh_inside(mesh1, mesh2) or self.is_mesh_inside(mesh2, mesh1)):
                        skip = True
                    if not skip and not self.do_bounds_overlap(mesh1, mesh2):
                        skip = True
                    if not skip and (mesh1.n_open_edges + mesh2.n_open_edges > 0):
                        skip = True
                        if mesh1.n_open_edges > 0:
                            update = 'Warning: unable to check {} for overlaps\n'.format(comp1['name'])
                            update += '-> {} has {} open edges.\n'.format(comp1['name'], mesh1.n_open_edges)
                            checked.append(comp1['id'])
                        else:
                            update = 'Warning: unable to check {} for overlaps\n'.format(comp2['name'])
                            update += '-> {} has {} open edges.\n'.format(comp2['name'], mesh2.n_open_edges)
                            checked.append(comp2['id'])
                        if progress_obj:
                            if progress_obj.sync_status(update=update): return
                        else:
                            print(update)

                    if not skip:
                        points, overlap_fraction = self.get_overlap(mesh1, mesh2, tolerance, n_samples, progress_obj)
                        threshold = n_samples * tolerance

                        if overlap_fraction is None:
                            update = 'Warning: insufficient sample points to check for overlap between {} and {}\n'\
                                     .format(comp1['name'], comp2['name'])
                            if progress_obj:
                                if progress_obj.sync_status(update=update): return
                            else:
                                print(update)
                        elif points.n_points > threshold:
                            overlapping_meshes.append(comp1['id'])
                            overlapping_meshes.append(comp2['id'])
                            actor = self.plotter.add_mesh(points, color='red', style='points', show_edges=False)
                            self.overlaps.append(actor)
                            update = 'Warning: {} may overlap {} by {:.3f} percent\n'\
                                    .format(comp1['name'], comp2['name'], 100.*overlap_fraction)
                            if progress_obj:
                                if progress_obj.sync_status(update=update): return
                            else:
                                print(update)

                if len(comp2['children']) > 0:
                    check_for_overlaps(comp1, comp2['children'], progress_obj)

        if progress_obj:
            progress_obj.reset_progress()
            num_components = self.count_components(self.components[0]['children'], exclude_events=True, \
                                                   exclude_invisible=True)
            num_checks = int(num_components*(num_components - 1)/2)
            progress_obj.set_maximum_value(num_checks)

        find_overlaps_recursive(self.components[0]['children'], progress_obj=progress_obj)

        if progress_obj:
            progress_obj.signal_finished()

        return np.unique(overlapping_meshes)


    def clear_component_meshes(self, components):
        """Clears the meshes in the components.

        :param components: The components to clear the meshes from.
        :type components: list
        """
        for comp in components:
            children = comp.pop('children', [])
            comp.clear()
            comp['children'] = children

            if 'mesh' in comp:
                comp['mesh'] = None

            if children:
                self.clear_component_meshes(children)
    
    
    def clear_meshes(self):
        """Clears the meshes and frees associated memory.
        """
        for actor in list(self.plotter.renderer.actors.values()):
            self.plotter.remove_actor(actor)
        
        self.actors.clear()
        self.clear_component_meshes(self.components)
        self.components.clear()
        self.overlaps.clear()
        self.event_ids.clear()
        self.num_to_plot = 0

        gc.collect()

        if not self.off_screen:
            self.plotter.update()



import os
import argparse
from pathlib import Path
from geviewer.gui import launch_app
from geviewer.utils import check_files


def main():
    """The main command-line entry point for GeViewer.
    """
    parser = argparse.ArgumentParser(
        description = 'GeViewer is a lightweight, Python-based visualization tool for Geant4.',
        epilog = 'For more information, visit https://geviewer.readthedocs.io/en/latest/'
    )
    parser.add_argument('files', nargs='*', help='Files to load on startup')
    args = parser.parse_args()

    if check_files(args.files):
        launch_app(args.files)
    

if __name__ == '__main__':
    main()



import os
from pathlib import Path


def check_files(files):
    """Checks if the file paths are valid.
    """
    for file in files:
        file = str(Path(file).resolve())
        if not os.path.exists(file):
            print('Error: {} does not exist'.format(file))
            return False
        if not (file.endswith('.gev') or file.endswith('.heprep') or file.endswith('.wrl')):
            print('Error: {} is not a valid file'.format(file))
            print('Valid file types are .gev, .heprep, and .wrl')
            return False
    return True


def check_for_updates():
    """Determines whether the user is using the latest version of GeViewer.
    If not, prints a message to the console to inform the user.
    """
    try:
        import json
        from urllib import request
        import geviewer
        from packaging.version import parse
        url = 'https://pypi.python.org/pypi/geviewer/json'
        releases = json.loads(request.urlopen(url).read())['releases']
        versions = list(releases.keys())
        parsed = [parse(v) for v in versions]
        latest = parsed[parsed.index(max(parsed))]
        current = parse(geviewer.__version__)
        if current < latest and not (latest.is_prerelease or latest.is_postrelease or latest.is_devrelease):
            msg = 'You are using GeViewer version {}. The latest version is {}. '.format(current, latest)
            msg += 'Use "pip install --upgrade geviewer" to update to the latest version.'
            return msg
        return
    except:
        # don't want this to interrupt regular use if there's a problem
        return


def print_banner():
    """Prints the banner to the terminal.
    """
    print()
    print('###################################################')
    print('#    _____   __      ___                          #')
    print('#   / ____|  \\ \\    / (_)                         #')
    print('#  | |  __  __\\ \\  / / _  _____      _____ _ __   #')
    print('#  | | |_ |/ _ \\ \\/ / | |/ _ \\ \\ /\\ / / _ \\  __|  #')
    print('#  | |__| |  __/\\  /  | |  __/\\ V  V /  __/ |     #')
    print('#   \\_____|\\___| \\/   |_|\\___| \\_/\\_/ \\___|_|     #')
    print('#                                                 #')
    print('###################################################')
    print()